#ifndef _SCORINGMATRIX_HPP_
#define _SCORINGMATRIX_HPP_

#include <iostream>
#include <stdexcept>
#include <string>
#include <vector>
#include <map>
#include <algorithm>

class ScoringMatrix {
public:
	//Constructor interfaces
	inline explicit ScoringMatrix(std::vector<std::string> alphabet);
	inline explicit ScoringMatrix();
	//copy constructor 
	inline ScoringMatrix(const ScoringMatrix& m);
	//copy operator
	inline ScoringMatrix& operator=(const ScoringMatrix& m);
	//Destructor interface
	inline ~ScoringMatrix();
	//+ operator
	inline ScoringMatrix& operator+=(const ScoringMatrix& m);
	inline ScoringMatrix operator+(const ScoringMatrix& m) const;
	//Java-style method interfaces.
	inline double get(std::string leftSymbol, std::string rightSymbol) const;
	inline double get(unsigned int leftSymbolIndex, unsigned int rightSymbolIndex) const;
	inline void set(std::string leftSymbol, std::string rightSymbol,
			double score);
	inline void set(unsigned int leftSymbolIndex, unsigned int rightSymbolIndex,
			double score);
	inline unsigned int size() const;
	inline std::string getSymbol(int index) const;
	inline int getIndex(std::string symbol) const;
	inline void setEmptyFlag();
	inline bool getEmptyFlag() const;

	inline bool isEmpty() const {
		return sz == 0;
	}

private:
	std::map<std::string, int> indexMapping;
	double ** scores;
	unsigned int sz;
	bool empty;
};

inline std::ostream& operator<<(std::ostream& os, const ScoringMatrix& m);
inline std::istream& operator>>(std::istream& is, ScoringMatrix& tr);

//constructor implementation

inline ScoringMatrix::ScoringMatrix() {
	sz = 0;
	scores = 0;
	empty = false;
}

inline ScoringMatrix::ScoringMatrix(std::vector<std::string> alphabet) {
	indexMapping = std::map<std::string, int>();
	unsigned int i = 0;
	unsigned int N = alphabet.size();
	std::string gap("-");
	std::string skip("_");
	for (std::vector<std::string>::iterator it = alphabet.begin(); it != alphabet.end(); ++it) {
		if (*it == gap || *it == skip) {
			N--;
			continue;
		}
		indexMapping.insert(std::pair<std::string, int>(*it, i++));
	}
	indexMapping.insert(std::pair<std::string, int>(gap, i++));
	indexMapping.insert(std::pair<std::string, int>(skip, i++));

	sz = i;
	scores = new double*[sz];
	for (i = 0; i < sz; i++) {
		scores[i] = new double[sz];
	}
	//initialize.
	for (i = 0; i < sz; i++) {
		for (unsigned int j = 0; j < sz; j++) {
			if (i < N && j < N) {
				if (i != j) {
					//mismatch cost.
					scores[i][j] = 2.;
				} else {
					//match cost.
					scores[i][j] = 0.;
				}
			} else if (i > N || j > N) {
				//skip cost.
				scores[i][j] = 1.;
			} else {
				//gap cost.
				scores[i][j] = 2.;
			}
		}
	}
	empty = false;
}

//copy constructor

inline ScoringMatrix::ScoringMatrix(const ScoringMatrix & m) {
	indexMapping = std::map<std::string, int>(m.indexMapping);
	sz = m.sz;
	scores = new double*[sz];
	for (unsigned int i = 0; i < sz; i++) {
		scores[i] = new double[sz];
	}
	for (unsigned int i = 0; i < sz; i++) {
		for (unsigned int j = 0; j < sz; j++) {
			scores[i][j] = m.scores[i][j];
		}
	}
	empty = false;
}
//copy operator
inline ScoringMatrix & ScoringMatrix::operator = (const ScoringMatrix & m)
{
	indexMapping = std::map<std::string, int>(m.indexMapping);
	double** tmp_scores = new double*[m.sz];
	for (unsigned int i = 0; i < m.sz; i++) {
		tmp_scores[i] = new double[m.sz];
	}
	for (unsigned int i = 0; i < m.sz; i++) {
		for (unsigned int j = 0; j < m.sz; j++) {
			tmp_scores[i][j] = m.scores[i][j];
		}
	}
	for (unsigned int i = 0; i < sz; i++) {
		delete[] scores[i];
	}
	delete[] scores;
	scores = tmp_scores;
	sz = m.sz;
	return *this;
}

//destructor

inline ScoringMatrix::~ScoringMatrix() {
	for (unsigned int i = 0; i < sz; i++) {
		delete[] scores[i];
	}
	delete[] scores;
}

inline unsigned int ScoringMatrix::size() const {
	return sz;
}


//+ operator
inline ScoringMatrix& ScoringMatrix::operator+=(const ScoringMatrix& m){
	assert(sz == m.size());
	
	for (unsigned int i = 0; i < m.size(); i++) {
		for (unsigned int j = 0; j < m.size(); j++) {
			scores[i][j] += m.get(i,j);
		}
	}

	return *this;
}

inline ScoringMatrix ScoringMatrix::operator+(const ScoringMatrix& m) const{
	assert(sz == m.size());
	//copy
	ScoringMatrix out = *this;
	out += m;
	return out;
}

//Getter and setter

inline double ScoringMatrix::get(std::string leftSymbol, std::string rightSymbol) const {
	int i = 0;
	try {
		i = indexMapping.at(leftSymbol);
	} catch (std::out_of_range) {
		std::string msg = "the symbol " + leftSymbol + " is unknown in the given scoring!";
		throw std::invalid_argument(msg.c_str());
	}
	int j = 0;
	try {
		j = indexMapping.at(rightSymbol);
	} catch (std::out_of_range) {
		std::string msg = "the symbol " + rightSymbol + " is unknown in the given scoring!";
		throw std::invalid_argument(msg.c_str());
	}
	return scores[i][j];
}

inline double ScoringMatrix::get(unsigned int leftSymbolIndex,
		unsigned int rightSymbolIndex) const {
	if (leftSymbolIndex >= sz || leftSymbolIndex < 0 ||
			rightSymbolIndex >= sz || rightSymbolIndex < 0) {
		throw std::out_of_range("The given indices were out of range!");
	}
	return scores[leftSymbolIndex][rightSymbolIndex];
}

inline void ScoringMatrix::set(std::string leftSymbol, std::string rightSymbol,
		double score) {
	int i = 0;
	try {
		i = indexMapping.at(leftSymbol);
	} catch (std::out_of_range) {
		std::string msg = "the symbol " + leftSymbol + " is unknown in the given scoring!";
		throw std::invalid_argument(msg.c_str());
	}
	int j = 0;
	try {
		j = indexMapping.at(rightSymbol);
	} catch (std::out_of_range) {
		std::string msg = "the symbol " + rightSymbol + " is unknown in the given scoring!";
		throw std::invalid_argument(msg.c_str());
	}
	scores[i][j] = score;
}

inline void ScoringMatrix::set(unsigned int leftSymbolIndex,
		unsigned int rightSymbolIndex, double score) {
	if (leftSymbolIndex >= sz || leftSymbolIndex < 0 ||
			rightSymbolIndex >= sz || rightSymbolIndex < 0) {
		throw std::out_of_range("The given indices were out of range!");
	}
	scores[leftSymbolIndex][rightSymbolIndex] = score;
}

struct Equals {
	int val;

	Equals(int n) {
		val = n;
	}

	inline bool operator()(const std::pair<std::string, int>& p) {
		return p.second == val;
	}
};

inline std::string ScoringMatrix::getSymbol(int index) const {

	std::_Rb_tree_const_iterator<std::pair<const std::basic_string<char>, int> > p =
			std::find_if(indexMapping.begin(), indexMapping.end(),
			Equals(index));

	return (*p).first;
}

inline int ScoringMatrix::getIndex(std::string symbol) const {
	return indexMapping.at(symbol);
}

inline void ScoringMatrix::setEmptyFlag() {
	empty = true;
}

inline bool ScoringMatrix::getEmptyFlag() const {
	return empty;
}

//global getters and setters


inline double get(const ScoringMatrix & matrix, char left, char right) {
	std::stringstream left_ss;
	std::string leftSymbol;
	left_ss << left;
	left_ss >> leftSymbol;
	std::stringstream right_ss;
	std::string rightSymbol;
	right_ss << right;
	right_ss >> rightSymbol;

	return matrix.get(leftSymbol,rightSymbol);
}

inline void set(ScoringMatrix & matrix, char left, char right, double score) {
	std::stringstream left_ss;
	std::string leftSymbol;
	left_ss << left;
	left_ss >> leftSymbol;
	std::stringstream right_ss;
	std::string rightSymbol;
	right_ss << right;
	right_ss >> rightSymbol;

	matrix.set(leftSymbol,rightSymbol, score);
}

//stream operator

inline std::ostream & operator<<(std::ostream& os, const ScoringMatrix & m) {
	os << "alphabet: ";
	for (unsigned int i = 0; i < m.size() - 2; i++) {
		os << "\"" << m.getSymbol(i) << "\",";
	}
	os << "\n";
	for (unsigned int i = 0; i < m.size(); i++) {
		for (unsigned int j = 0; j < m.size(); j++) {
			if (i < m.size() - 2 || j < m.size() - 2) {
				os << "\"" << m.getSymbol(i) << "\"->\""
						<< m.getSymbol(j) << "\":" <<
						m.get(i, j) << "\n";
			}
		}
	}
	return os;
}

inline std::istream & operator>>(std::istream& is, ScoringMatrix & m) {
	std::string prefix;
	is >> prefix;
	std::string expected_prefix("alphabet:");
	if (prefix != expected_prefix) {
		throw std::invalid_argument("prefix was wrong");
		is.setstate(std::ios::failbit);
		return is;
	}
	char c;
	std::vector<std::string> alphabet;
	while (is.peek() != '\n') {
		std::string symbol;
		is >> c;
		if (c != '\"') {
			throw std::invalid_argument("expected \" before beginning of alphabet symbol");
			is.setstate(std::ios::failbit);
			return is;
		}
		while (is >> c && c != '\"') {
			symbol += c;
		}
		is >> c;
		if (c != ',') {
			throw std::invalid_argument("expected , after alphabet symbol");
			is.setstate(std::ios::failbit);
			return is;
		}
		alphabet.push_back(symbol);
	}
	m = ScoringMatrix(alphabet);
	while (is >> c) {
		if (c != '\"') {
			throw std::invalid_argument("expected \" before beginning of first scoring matrix argument");
			is.setstate(std::ios::failbit);
			return is;
		}
		std::string leftSym;
		while (is >> c && c != '\"') {
			leftSym += c;
		}
		is >> c;
		if (c != '-') {
			throw std::invalid_argument("expected arrow (->) between scoring matrix arguments");
			is.setstate(std::ios::failbit);
			return is;
		}
		is >> c;
		if (c != '>') {
			throw std::invalid_argument("expected arrow (->) between scoring matrix arguments");
			is.setstate(std::ios::failbit);
			return is;
		}
		is >> c;
		if (c != '\"') {
			throw std::invalid_argument("expected \" before beginning of scond scoring matrix argument");
			is.setstate(std::ios::failbit);
			return is;
		}
		std::string rightSym;
		while (is >> c && c != '\"') {
			rightSym += c;
		}
		is >> c;
		if (c != ':') {
			throw std::invalid_argument("expected : between scoring matrix arguments and score");
			is.setstate(std::ios::failbit);
			return is;
		}
		double score;
		is >> score;
		m.set(leftSym, rightSym, score);
	}
	return is;
}

//gapc empty commands
inline void empty(ScoringMatrix &m) {m.setEmptyFlag(); }
inline bool isEmpty(const ScoringMatrix &m) { return m.getEmptyFlag(); }

#endif
