/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This program is written in GAP-L as developed by Georg Sauthoff at Bielefeld
 * University. GAP-L, the Bellman's Gap Compiler and other language features are
 * used in accordance to the GPL v3+.
 *
 * Bellman's Gap can be found at http://www.gapc.eu
 *
 * Special thanks goes to Robert Giegerich, Stefan Janssen and Benedikt Löwes
 * for their support with Bellman's Gap.
 */

/**
 * This implements the local sequence alignment on simple character sequences.
 * Every character is interpreted as a discrete symbol.
 *
 * @author =  Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
//just some preamble definitions

import align_cost
import twoline_Rope
import limit_id
import softmin
import ScoringMatrix
import soft_derivative
import softmin_deriv

input < raw, raw >

type Rope = extern
type twoline_Rope = extern
type ScoringMatrix = extern
type soft_derivative = extern

/*
 * Signature that defines the operations allowed.
 */
signature sig_alignment(alphabet, answer) {
  answer del(<alphabet,void>,answer);
  answer skip_del(<alphabet,void>,answer);
  answer ins(<void,alphabet>,answer);
  answer skip_ins(<void,alphabet>,answer);
  answer rep(<alphabet,alphabet>,answer);
  answer nil(<void,void>);
  choice [answer] h([answer]);
}

/*
 * Algebra section.
 */
algebra enum auto enum ;
algebra count auto count;

//this implements the strict or classical alignment scoring.
algebra alg_strict implements sig_alignment(alphabet = char, answer = double){

	double del(<char left, void>, double in){
		return in + del_cost(left);
	}
	
	double skip_del(<char left, void>, double in){
		return in + skip_del_cost(left);
	}
	
	double ins(<void, char right>, double in){
		return in + ins_cost(right);
	}

	double skip_ins(<void, char right>, double in){
		return in + skip_ins_cost(right);
	}

	double rep(<char left, char right>, double in){
		return in + rep_cost(left, right);
	}
	
	double nil(<void, void>){
		return 0;
	}
	
	choice [double] h([double] l){
		return list(minimum(l));
	}
}

//this replaces the strict minimum by softmin.
algebra alg_soft extends alg_strict {

	scoring choice [double] h([double] l){
		return list(softmin(l));
	}

}

/*
 * here we calculate a strict derivative for the given input sequences
 * regarding all parameters. It basically just counts the number of times a
 * given parameter has been used and stores that in a scoring matrix.
 *
 * Note that this only calculates the derivative if in the product
 * alg_strict * alg_deriv
 *
 * There is a problem, however, with suboptimal solutions. Here the outcome
 * depends on how you "define" the derivative of the (in principle) non-derivable
 * strict min.
 * Assume we have min(x_1,...,x_n) and let I be the set of indices of co-optimals,
 * meaning that for all i in I: x_i = min(x_1, ... , x_n).
 * Then we have three possibilities:
 * 1.) min'(x_i) := 1 for only one (randomly chosen) i in I.
 * 2.) min'(x_i) := 1/|I| for all i. This leads to an average over co-optimals
 * and is the "cleanest" way to some extend.
 * 3.) min'(x_i) := 1 for all i. This leads to a sum over all co-optimals.
 * It is somewhat bad because more co-optimal paths lead to a higher derivative
 * (which is counter-intuitive) but we chose that possibility here for reasons
 * of simplicity. To get a normalized number you should use the "num_coopts"
 * instance for your example and divide your result by the number of cooptimals.
 */
algebra alg_deriv implements sig_alignment(alphabet = char, answer = ScoringMatrix){

	ScoringMatrix del(<char left, void>, ScoringMatrix in){
		ScoringMatrix out = in;
		double old_param_deriv = get(out,left,'-');
		set(out,left,'-',old_param_deriv + 1);
		return out;
	}

	ScoringMatrix skip_del(<char left, void>, ScoringMatrix in){
		ScoringMatrix out = in;
		double old_param_deriv = get(out,left,'_');
		set(out,left,'_',old_param_deriv + 1);
		return out;
	}

	ScoringMatrix ins(<void, char right>, ScoringMatrix in){
		ScoringMatrix out = in;
		double old_param_deriv = get(out,'-',right);
		set(out,'-',right,old_param_deriv + 1);
		return out;
	}

	ScoringMatrix skip_ins(<void, char right>, ScoringMatrix in){
		ScoringMatrix out = in;
		double old_param_deriv = get(out,'_',right);
		set(out,'_',right,old_param_deriv + 1);
		return out;
	}

	ScoringMatrix rep(<char left, char right>, ScoringMatrix in){
		ScoringMatrix out = in;
		double old_param_deriv = get(out,left,right);
		set(out,left,right,old_param_deriv + 1);
		return out;
	}
	
	ScoringMatrix nil(<void, void>){
		ScoringMatrix init = getInitial();
		return init;
	}
	
	choice [ScoringMatrix] h([ScoringMatrix] l){
		return list(sum(l));
	}
}

/*
 * The soft alignment replaces the minimum mentioned above in the strict derivative
 * by the softmin function, which is differentiable. Unfortunately, though,
 * it is not possible anymore to get the derivative by an algebra product.
 * We do have to calculate everything ourselves now. For more details on the
 * derivative of soft alignment please have a look at our papers.
 */
algebra alg_soft_deriv implements sig_alignment(alphabet = char, answer = soft_derivative){

	soft_derivative del(<char left, void>, soft_derivative in){
		soft_derivative out;
		out.softmin = in.softmin + del_cost(left);
		out.deriv = in.deriv;
		double old_param_deriv = get(in.deriv,left,'-');
		set(out.deriv,left,'-',old_param_deriv + 1);
		return out;
	}

	soft_derivative skip_del(<char left, void>, soft_derivative in){
		soft_derivative out;
		out.softmin = in.softmin + skip_del_cost(left);
		out.deriv = in.deriv;
		double old_param_deriv = get(in.deriv,left,'_');
		set(out.deriv,left,'_',old_param_deriv + 1);
		return out;
	}

	soft_derivative ins(<void, char right>, soft_derivative in){
		soft_derivative out;
		out.softmin = in.softmin + ins_cost(right);
		out.deriv = in.deriv;
		double old_param_deriv = get(in.deriv,'-',right);
		set(out.deriv,'-',right,old_param_deriv + 1);
		return out;
	}

	soft_derivative skip_ins(<void, char right>, soft_derivative in){
		soft_derivative out;
		out.softmin = in.softmin + skip_ins_cost(right);
		out.deriv = in.deriv;
		double old_param_deriv = get(in.deriv,'_',right);
		set(out.deriv,'_',right,old_param_deriv + 1);
		return out;
	}

	soft_derivative rep(<char left, char right>, soft_derivative in){
		soft_derivative out;
		out.softmin = in.softmin + rep_cost(left,right);
		out.deriv = in.deriv;
		double old_param_deriv = get(in.deriv,left,right);
		set(out.deriv,left,right,old_param_deriv + 1);
		return out;
	}
	
	soft_derivative nil(<void, void>){
		soft_derivative init;
		init.deriv = getInitial();
		init.softmin = 0;
		return init;
	}
	
	scoring choice [soft_derivative] h([soft_derivative] l){
		return list(softmin_deriv(l));
	}
}

//This is a helper algebra to minimize the number of gaps and maximize the
//number of replacements.
algebra alg_min_gaps implements sig_alignment(alphabet = char, answer = int){

	int del(<char left, void>, int in){
		return in + 1;
	}
	
	int skip_del(<char left, void>, int in){
		return in + 1;
	}
	
	int ins(<void, char right>, int in){
		return in + 1;
	}

	int skip_ins(<void, char right>, int in){
		return in + 1;
	}

	int rep(<char left, char right>, int in){
		return in;
	}
	
	int nil(<void, void>){
		return 0;
	}
	
	choice [int] h([int] l){
		return list(minimum(l));
	}
}

//A printing algebra.
algebra alg_print implements sig_alignment(alphabet = char, answer = twoline_Rope){

	twoline_Rope del(<char left,void>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, left);
		append(out.top, in.top);
		append(out.bottom, '-');
		append(out.bottom, in.bottom);
		return out;
	}

	twoline_Rope skip_del(<char left,void>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, left);
		append(out.top, in.top);
		append(out.bottom, '_');
		append(out.bottom, in.bottom);
		return out;
	}

	twoline_Rope ins(<void,char right>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, '-');
		append(out.top, in.top);
		append(out.bottom, right);
		append(out.bottom, in.bottom);
		return out;
	}
	
	twoline_Rope skip_ins(<void,char right>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, '_');
		append(out.top, in.top);
		append(out.bottom, right);
		append(out.bottom, in.bottom);
		return out;
	}
	
	twoline_Rope rep(<char left, char right>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, left);
		append(out.top, in.top);
		append(out.bottom, right);
		append(out.bottom, in.bottom);
		return out;
	}
	
	twoline_Rope nil(<void,void>){
		twoline_Rope out;
		return out;
	}
	
	choice [twoline_Rope] h([twoline_Rope] l){
		return l;
	}
}

//A printing algebra again, but limiting the number of paths that are printed.
algebra alg_limit_print extends alg_print {

	kscoring choice [twoline_Rope] h([twoline_Rope] l){
		return limit_id(l);
	}

}

/*
 * Grammar to define what kind of alignments are allowed and which are not.
 * Note that we take the following precautions:
 * - We enforce that at least one replacement happens before we can start
 *   skipping symbols at the end.
 * - We enforce that deletions always are in front of insertions.
 * - We enforce that skip regions in the middle have at least length 4 to
 *   prevent trivial skip regions.
 */
grammar gra_alignment uses sig_alignment(axiom = SKIPDEL_START)
{

	//skip the start
	SKIPDEL_START = skip_del(<CHAR,EMPTY>, SKIPDEL_START) | SKIPINS_START #h;
	
	SKIPINS_START = skip_ins(<EMPTY,CHAR>, SKIPINS_START) |
					rep(<CHAR,CHAR>, ALI) |
					nil(<EMPTY,EMPTY>) #h;

	//align in the middle
	ALI = del(<CHAR,EMPTY>, DEL) | ins(<EMPTY,CHAR>, INS) | rep(<CHAR,CHAR>, ALI) |
			SKIPDEL_MIDDLE | SKIPINS_MIDDLE | SKIPDEL_END #h;
	
	DEL = del(<CHAR,EMPTY>, DEL) | ins(<EMPTY,CHAR>, INS) | rep(<CHAR,CHAR>, ALI) #h;

	INS = ins(<EMPTY,CHAR>, INS) | rep(<CHAR,CHAR>, ALI) #h;
	
	//skip to a next interesting region with at least 4 skips
	SKIPDEL_MIDDLE = skip_del(<CHAR,EMPTY>,
						skip_del(<CHAR,EMPTY>,
						skip_del(<CHAR,EMPTY>,
						skip_del(<CHAR,EMPTY>,
						SKIPDEL_MIDDLE_LOOP)))) #h;
	
	SKIPDEL_MIDDLE_LOOP = skip_del(<CHAR,EMPTY>, SKIPDEL_MIDDLE_LOOP) |
							rep(<CHAR,CHAR>,ALI) #h;
	
	SKIPINS_MIDDLE = skip_ins(<EMPTY,CHAR>,
						skip_ins(<EMPTY,CHAR>,
						skip_ins(<EMPTY,CHAR>,
						skip_ins(<EMPTY,CHAR>,
						SKIPINS_MIDDLE_LOOP)))) #h;

	SKIPINS_MIDDLE_LOOP = skip_ins(<EMPTY,CHAR>, SKIPINS_MIDDLE_LOOP) |
							rep(<CHAR,CHAR>,ALI) #h;

	//skip the end
	SKIPDEL_END = skip_del(<CHAR,EMPTY>, SKIPDEL_END) | SKIPINS_END #h;

	SKIPINS_END = skip_ins(<EMPTY,CHAR>, SKIPINS_END) | nil(<EMPTY,EMPTY>) #h;
}


/*
 * This defines compilable instances of this program.
 * The command line usage (if gapc is installed):
 *
 * gapc -i <instance name> -o local_alignment.cc local_alignment.gap
 *
 */
instance print		= gra_alignment(alg_print);
instance score		= gra_alignment(alg_strict);
instance soft_score	= gra_alignment(alg_soft);
instance deriv		= gra_alignment(alg_strict * alg_deriv);
instance soft_deriv	= gra_alignment(alg_soft_deriv);
instance num_coopts	= gra_alignment(alg_strict * count);
instance path		= gra_alignment(alg_strict * alg_min_gaps * alg_limit_print);
instance debug		= gra_alignment(alg_print * alg_strict);
