% TCS Alignment Toolbox
%
% Copyright (C) 2013, 2014
% Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
% AG Theoretical Computer Science
% Centre of Excellence Cognitive Interaction Technology (CITEC)
% University of Bielefeld
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
% 
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

% --- Sine Waves Dynamic Time Warping ---
%
% In this example we compare two sine wave sequences using dynamic
% time warping.

%% Preamble
%
% This includes the necessary classes from the TCSAlignmentToolbox
% .jar-file.


clear java;
clear;
clc;

javaaddpath TCSAlignmentToolbox-1.4.jar;
import de.unibi.citec.tcs.alignment.sequence.*;
import de.unibi.citec.tcs.alignment.*;
import de.unibi.citec.tcs.alignment.comparators.*;
import de.unibi.citec.tcs.alignment.wrappers.VectorialSequences;

%%  1.) Create both sine waves. One with doubled frequency.
data = zeros(2,40);
data(1,:) = sin((pi * 0.1) * (1:40));
data(2,:) = sin((pi * 0.05) * (1:40));

% plot it
figure(1);
plot(1:40,data(1,:),1:40,data(2,:));
legend('wave 1', 'wave 2');
title('Two example sine waves');

%% 2.) Transform to sequences.
sequences = VectorialSequences.toSequences(data);
% plot one sequence to be sure that the conversion worked.
sequences(1).toString()
%% 3.) Create AlignmentSpecification using a default comparator.
alignSpec = VectorialSequences.setUpSpecification(sequences);
%  show it.
alignSpec.toString()
%  Maybe we want to use the L2 norm in this case rather than L1.
comp = EuclideanComparator(1);
alignSpec = VectorialSequences.setUpSpecification(sequences,comp);

%% 4.) Compare both sequences using Dynamic Time Warping.
dtwPathAlgo = StrictDTWFullAlgorithm(alignSpec);
path = dtwPathAlgo.calculateAlignment(sequences(1), sequences(2));
% These have fairly low distance because DTW can account for different frequencies.
path.getScore()
% also show the extensions and the difference function.
exts = VectorialSequences.getExtensions(path);
diff = VectorialSequences.getDifference(path);

ext_length = numel(diff);

figure(2);
plot(1:ext_length,exts(1,:),1:ext_length,exts(2,:),1:ext_length,diff);
legend('wave 1 DTW extension', 'wave 2 DTW extension', 'DTW difference');
title('Dynamic Time Warping of both sine waves');

%% 5.) Now assume we have two classes of multi-dimensional waves.
%  One class has a slight linear slope in the first dimension, the
%  other one hasn't. That is the discriminative part. However the
%  other dimension is very noisy, which will obscure the discriminative
%  effect under default conditions.
%  create that data set.
numSeqs = 10;
seqLength = 100;
data = zeros(numSeqs,seqLength,2);
slope = 2 / seqLength;
% shortcut for the class size.
C = numSeqs / 2;
for s = 1:numSeqs
    % create a nice sine wave in the first dimension with slightly
    % noisy frequencies
    f = randn(1) * 0.01 + 0.1;
    data(s,:,1) = sin(pi * f * (1:seqLength));
    % add the slope if we are in class two.
    if (s > C) 
        data(s,:,1) = data(s,:,1) + ...
            ((1:seqLength) - ones(1,seqLength) * seqLength * 0.5) * slope;
    end
    % put random noise in the second dimension.
    data(s,:,2) = randn(seqLength,1);
end

% plot one sequence as an example.
figure(3);
plot(1:seqLength,data(C+1,:,1),1:seqLength,data(C+1,:,2));
legend('discriminative dimension','noise dimension');
title('one example sequence');

%% 6.) give a name to our dimensions and create the sequences.
keywords = javaArray('java.lang.String',2);
keywords(1) = java.lang.String('relevant');
keywords(2) = java.lang.String('noise');
sequences = VectorialSequences.toSequences(data, keywords);

%% 7.) Create a AlignmentSpecification
alignSpec = VectorialSequences.setUpSpecification(sequences)
%% 8.) Let's look at the score matrix.
dtwScoreAlgo = StrictDTWScoreAlgorithm(alignSpec);
parallelScoreEngine = ParallelProcessingEngine(dtwScoreAlgo, sequences);
parallelScoreEngine.calculate();
distMat = parallelScoreEngine.getScoreMatrix();

figure(4);
imagesc(distMat);
colorbar;
title('distance matrix');

%% 9.) This obsiously is not perfect yet.
%  So let's take a closer look at the data space. Using the derivative
%  we can abstract from the current weighting scheme and look which
%  dimension has especially high intra- or inter-class distance. To
%  improve class-discrimination we then can lower the weight on a
%  dimension that has higher (or equal) intra-class distance than inter-
%  class-distance and we can emphasize dimensions with higher inter-
%  class-distance that actually help us to discriminate.
%  this is a very easy learning scheme inspired by the GLVQ cost function.

% first calculate full paths to have enough information for the derivative..
dtwPathAlgo = StrictDTWFullAlgorithm(alignSpec);
parallelPathEngine = ParallelProcessingEngine(dtwPathAlgo, sequences);
parallelPathEngine.calculate();
paths = parallelPathEngine.getResultMatrix();

% then calculate the weight derivative.
derivEngine = ParallelDerivativeEngine(paths);
weightDerivs = derivEngine.calculateWeightDerivatives();

% then calculate the GLVQ gradient as sketched above.
deltaW = zeros(2,1);

k = 0;
for i = 1:numSeqs
    for j = 1:numSeqs
        k = k+1;
        for w = 1:2
            if ((i < C && j < C) || (i >= C && j >= C))
                %  intra class derivatives are added
                deltaW(w) = deltaW(w) + weightDerivs(k,w);
            else
                % inter class derivatives are subtracted
                deltaW(w) = deltaW(w) - weightDerivs(k,w);
            end
        end
    end
end

% show the result.
deltaW

%%  10.) So we should emphasize the first dimension more than the second
%  one. Let's do that and look at the matrix again.
alignSpec.setWeighting([0.7, 0.3]);
dtwScoreAlgo = StrictDTWScoreAlgorithm(alignSpec);
parallelScoreEngine = ParallelProcessingEngine(dtwScoreAlgo, sequences);
parallelScoreEngine.calculate();
distMat = parallelScoreEngine.getScoreMatrix();

figure(5);
imagesc(distMat);
colorbar;
title('distance matrix');
%%  11.) Better but still not perfect. Let's push it to the extreme.
alignSpec.setWeighting([1, 0.0]);
parallelScoreEngine.calculate();
distMat = parallelScoreEngine.getScoreMatrix();

figure(5);
imagesc(distMat);
colorbar;
title('distance matrix');
% now we have near-perfect class separation.