/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import de.unibi.citec.tcs.alignment.AlignmentPath;
import de.unibi.citec.tcs.alignment.AlignmentSpecification;
import de.unibi.citec.tcs.alignment.ParallelProcessingEngine;
import de.unibi.citec.tcs.alignment.SoftLocalAlignmentFullAlgorithm;
import de.unibi.citec.tcs.alignment.StrictAlignmentFullAlgorithm;
import de.unibi.citec.tcs.alignment.StrictAlignmentScoreAlgorithm;
import de.unibi.citec.tcs.alignment.StrictAllOptimalAlgorithm;
import de.unibi.citec.tcs.alignment.StrictLocalAlignmentFullAlgorithm;
import de.unibi.citec.tcs.alignment.StrictLocalAlignmentScoreAlgorithm;
import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.wrappers.StringEditDistance;

/**
 *
 * --- String Edit Distance Example ---
 *
 *
 * In this example we compare multiple strings in a simple
 * StringEditDistance fashion. We also apply a learning scheme to increase
 * class separation using metric learning only.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class string_edit_distance {

	public static void main(String[] args) {
		// 1.) Set up your data space. In this case we assume to have two
		// distinct classes of strings that only consists of "a"s and "b". The only
		// difference between both classes is, that the members of the second one
		// are longer.

		final String[] strings = new String[6];

		// first class with shorter sequences.
		strings[0] = "ababababa";
		strings[1] = "babababab";
		strings[2] = "bababababb";
		// second class with longer sequences.
		strings[3] = "ababababababbbb";
		strings[4] = "bababbababababab";
		strings[5] = "aabababaaabbabab";

		// 2.) convert to the TCSAlignmentToolbox sequence format.
		final Sequence[] sequences = StringEditDistance.toSequences(strings);

		// show the first sequence again for testing purposes
		System.out.println("First Sequence");
		System.out.println(sequences[0].toString());
		System.out.println("");

		// 3.) Set up the default edit distance scoring scheme.
		// Matches cost nothing, everything else costs 1.
		AlignmentSpecification alignSpec
				= StringEditDistance.setUpSpecification(sequences);

		// 4.) Set up your algorithm of choice.
		final StrictAlignmentFullAlgorithm algo
				= new StrictAlignmentFullAlgorithm(alignSpec);
		// show one example alignment to see, if our scoring scheme works.
		final AlignmentPath alignmentPath
				= algo.calculateAlignment(sequences[0], sequences[3]);
		System.out.println("Alignment Path (inter class)");
		System.out.println(alignmentPath);
		System.out.println("");

		// 5.) calculate the distance matrix in our data space in parallel.
		StrictAlignmentScoreAlgorithm scoreAlgo
				= new StrictAlignmentScoreAlgorithm(alignSpec);

		ParallelProcessingEngine parallelEngine
				= new ParallelProcessingEngine(scoreAlgo, sequences);
		parallelEngine.setNumberOfThreads(32);
		parallelEngine.calculate();
		double[][] scoreMat = parallelEngine.getScoreMatrix();

		// plot the matrix
		System.out.println("\nScore Matrix Pre-Learning\n");
		plotScoreMatrix(scoreMat);
		System.out.println("");

		// this separates the classes already quite well, but it is not yet optimal.
		// 6.) calculate the derivative.
		// we observe, if we look at the derivative inside a class and between
		// classes, that more deletions and insertions are used between classes.
		final StrictAllOptimalAlgorithm derivAlgo
				= new StrictAllOptimalAlgorithm(alignSpec);

		double[][] deriv = StringEditDistance.calculateDerivative(
				derivAlgo, sequences[0], sequences[1]);

		// plot the matrix of parameter derivatives for an intra-class alignment an
		// an inter-class alignment. Inside the class we observe a higher
		// replacement derivative (the section (1,1)-(2,2) of the matrix, where the
		// diagonal models replacements of equal symbols).
		// Between classes we observe more insertions (becuse the second class has
		// longer sequences; the section (3,1)-(3,2) of the matrix, where (3,1) is
		// an insertion of a and (3,2) is an insertion of b).
		System.out.println("\nDerivative matrix (intra class)\n");
		plotDerivMatrix(deriv, "ab".toCharArray(), false);
		System.out.println("");

		deriv = StringEditDistance.calculateDerivative(
				derivAlgo, sequences[0], sequences[3]);

		System.out.println("\nDerivative matrix (inter class)\n");
		plotDerivMatrix(deriv, "ab".toCharArray(), false);
		System.out.println("");

		// 7.) let's check if that observation holds for the entire data space.
		// the last parameter here is the number of threads for parallel
		// computation.
		double[][][][] derivs = StringEditDistance.calculateDerivatives(
				derivAlgo, sequences, 32);

		// this methods returns _all_ pairwise derivatives. To get an overview of
		// the overall derivative differences we add all intra-class-derivatives and
		// subtract all inter-class-derivatives.
		deriv = new double[3][3];

		for (int i = 0; i < 6; i++) {
			for (int j = 0; j < 6; j++) {
				for (int k = 0; k < 3; k++) {
					for (int m = 0; m < 3; m++) {
						if ((i < 3 && j < 3) || (i >= 3 && j >= 3)) {
							// intra class derivatives are added
							deriv[k][m] += derivs[i][j][k][m];
						} else {
							//inter class derivatives are subtracted
							deriv[k][m] -= derivs[i][j][k][m];
						}
					}
				}
			}
		}

		// plot the resulting matrix
		System.out.println("\nDerivative matrix (whole dataset)\n");
		plotDerivMatrix(deriv, "ab".toCharArray(), false);
		System.out.println("");

		// This resulting matrix actually corresponds to a (very simple) gradient on
		// the GLVQ cost function: To improve class separation we have to decrease
		// replacement costs (positive matrix values) and have to increase gap costs
		// (strongly negative matrix values).
		// 8.) set up new alignment specification according to the
		// observation. We use zero replacement costs and maximum (1) gap costs.
		// the second parameter is the cost for matches, the second one for
		// mismatches and the third one for gaps.
		alignSpec = StringEditDistance.setUpSpecification(sequences, 0, 0, 1);

		// calculate the distance matrix in our data space again.
		scoreAlgo = new StrictAlignmentScoreAlgorithm(alignSpec);

		parallelEngine = new ParallelProcessingEngine(scoreAlgo, sequences);
		parallelEngine.setNumberOfThreads(32);
		parallelEngine.calculate();
		scoreMat = parallelEngine.getScoreMatrix();

		// plot the matrix
		System.out.println("\nScore Matrix Post-Learning\n");
		plotScoreMatrix(scoreMat);
		System.out.println("");

		// now we have near perfect class separation.
		/*
		 * 9.) Now assume that we have a slightly different data space where
		 * the strings have strong local similarity but a lot of noise in
		 * between that is not important for our alignment. Like so.
		 */
		final String[] noisyStrings = new String[6];
		//class 1 with shorter (relevant) sequences.
		noisyStrings[0] = "zzzzzzzzzzaaaazzzzzzzzaaaaazzzzzzzzzzz";
		noisyStrings[1] = "yyyyyaaaaayyyyyyyyyyyyyyaaaayyyy";
		noisyStrings[2] = "xxxxxxxxxxxxxaaaxxxxyaaaaaxxxxxxxxxxxxxx";
		//class 2 with longer (relevant) sequences.
		noisyStrings[3] = "xxxxaaaaaaaaaxxxxaaaaaaaaaaaaaxxxxxxxxx";
		noisyStrings[4] = "yyyyyyyyyyaaaaaaaaaayyyyyyyyyaaaaaaayyyy";
		noisyStrings[5] = "zzzzzzzzzzzzzzzzzzaaaaaaaaazzzzzzzzaaaaaaaaaaaazzz";
		final Sequence[] noisySeqs = StringEditDistance.toSequences(noisyStrings);
		/*
		 * 10.) It that case it is smarter to use a Local (and affine) Alignment
		 * like so:
		 */
		alignSpec = StringEditDistance.setUpSpecification(noisySeqs);
		final StrictLocalAlignmentFullAlgorithm localAlgo
				= new StrictLocalAlignmentFullAlgorithm(alignSpec);
		final AlignmentPath localAlignmentPath
				= localAlgo.calculateAlignment(noisySeqs[0], noisySeqs[1]);
		System.out.println(localAlignmentPath);
		/*
		 * 11.) Let's have a look at our class separation.
		 */
		final StrictLocalAlignmentScoreAlgorithm localScoreAlgo
				= new StrictLocalAlignmentScoreAlgorithm(alignSpec);
		parallelEngine = new ParallelProcessingEngine(localScoreAlgo, noisySeqs);
		parallelEngine.calculate();
		scoreMat = parallelEngine.getScoreMatrix();
		System.out.println("\nScore Matrix Pre-Learning (Noisy Data Space)\n");
		plotScoreMatrix(scoreMat);
		System.out.println("");

		/*
		 * 12.) Now obviously in this data space it is best to set the costs
		 * for skipping longer, irrelevant and non-discriminative parts of the
		 * alignment to zero while we want to emphasize mismatch, deletion and
		 * skip costs for the symbol "a".
		 *
		 * Let's see if we can learn that from out data using a soft alignment
		 * approach.
		 */
		final SoftLocalAlignmentFullAlgorithm localDerivAlgo
				= new SoftLocalAlignmentFullAlgorithm(alignSpec);
		localDerivAlgo.setBeta(3);
		derivs = StringEditDistance.calculateDerivatives(localDerivAlgo, noisySeqs, 32);
		// As before we add all intra-class-derivatives and
		// subtract all inter-class-derivatives.
		final int A = 4 + 2;
		deriv = new double[A][A];

		for (int i = 0; i < 6; i++) {
			for (int j = 0; j < 6; j++) {
				for (int k = 0; k < A; k++) {
					for (int m = 0; m < A; m++) {
						if ((i < 3 && j < 3) || (i >= 3 && j >= 3)) {
							// intra class derivatives are added
							deriv[k][m] += derivs[i][j][k][m];
						} else {
							//inter class derivatives are subtracted
							deriv[k][m] -= derivs[i][j][k][m];
						}
					}
				}
			}
		}

		// plot the resulting matrix
		System.out.println("\nDerivative matrix (whole dataset)\n");
		plotDerivMatrix(deriv, "axyz".toCharArray(), true);
		System.out.println("");

		/*
		 * 13.) It is not as clear as in the previous example, but still the
		 * tendency is visible: While skip costs for "a"s should be made far
		 * more expensive, other skip costs should be made cheaper or should
		 * at least be much cheaper compared to "a" skips. Let's do that.
		 */
		final ReplacementComparator comparator
				= (ReplacementComparator) alignSpec.getComparator(0);
		comparator.getCostMatrix().setSkipDeletionCost("a", 1);
		comparator.getCostMatrix().setSkipInsertionCost("a", 1);
		comparator.getCostMatrix().setSkipDeletionCost("x", 0);
		comparator.getCostMatrix().setSkipInsertionCost("x", 0);
		comparator.getCostMatrix().setSkipDeletionCost("y", 0);
		comparator.getCostMatrix().setSkipInsertionCost("y", 0);
		comparator.getCostMatrix().setSkipDeletionCost("z", 0);
		comparator.getCostMatrix().setSkipInsertionCost("z", 0);
		//and have a look at the distance matrix again.
		parallelEngine.calculate();
		scoreMat = parallelEngine.getScoreMatrix();
		System.out.println("\nScore Matrix Post-Learning (Noisy Data Space)\n");
		plotScoreMatrix(scoreMat);
		System.out.println("");
		//again we achieved near-perfect class separation.
	}

	private static final int DECIMAL_PLACES = 2;

	private static void plotScoreMatrix(final double[][] scoreMat) {
		for (int i = 0; i < scoreMat.length; i++) {
			if (i == 0) {
				//add column labels
				System.out.print("  | ");
				for (int j = 0; j < scoreMat[i].length; j++) {
					System.out.print(j);
					for (int c = 0; c < DECIMAL_PLACES; c++) {
						System.out.print(" ");
					}
					System.out.print("| ");
				}
				System.out.println("");
			}
			//add current row label.
			System.out.print(i + " |");
			for (int j = 0; j < scoreMat[i].length; j++) {
				//add actual content.
				String contentStr = Double.toString(scoreMat[i][j]);
				if (contentStr.length() > DECIMAL_PLACES + 2) {
					contentStr = contentStr.substring(0, 4);
				}
				System.out.print(contentStr);
				//fill up missing spaces if necessary.
				for (int c = contentStr.length(); c < DECIMAL_PLACES + 2; c++) {
					System.out.print(" ");
				}
				System.out.print("|");
			}
			System.out.println("");
		}
	}

	private static void plotDerivMatrix(final double[][] deriv,
			char[] symbols, boolean withSkips) {
		int offset = 1;
		if (withSkips) {
			offset++;
		}
		for (int i = 0; i < symbols.length + offset; i++) {
			if (i == 0) {
				//add column labels
				System.out.print("  |  ");
				for (final char c : symbols) {
					System.out.print(c + " |  ");
				}
				System.out.print("-  ");
				if (withSkips) {
					System.out.print("|  _ ");
				}
				System.out.println("");
			}
			//add current row label.
			if (i == symbols.length + 1) {
				System.out.print("_ |");
			} else if (i == symbols.length) {
				System.out.print("- |");
			} else {
				System.out.print(symbols[i] + " |");
			}
			for (int j = 0; j < symbols.length + offset; j++) {
				//add actual content.
				String contentStr = Double.toString(deriv[i][j]);
				if (contentStr.length() > DECIMAL_PLACES + 2) {
					contentStr = contentStr.substring(0, 4);
				}
				System.out.print(contentStr);
				//fill up missing spaces if necessary.
				for (int c = contentStr.length(); c < DECIMAL_PLACES + 2; c++) {
					System.out.print(" ");
				}
				System.out.print("|");
			}
			System.out.println("");
		}
	}
}
