/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.DerivableComparator;
import de.unibi.citec.tcs.alignment.sequence.Value;

/**
 * This is an interface to provide calculation methods for the derivative of an
 * alignment w.r.t. a specific comparator and the weights for the keywords.
 *
 * Please note that this is an interface to be implemented by the results of
 * AlignmentAlgorithms, not the algorithms itself. For example the result of the
 * StrictAlignmentFullAlgorithm is an AlignmentPath, which is an
 * AlignmentDerivativeAlgorithm, because the AlignmentPath contains the
 * necessary information to calculate the derivative.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public interface AlignmentDerivativeAlgorithm {

	/**
	 * This returns the derivatives for all parameters of the given comparator
	 * according to the implementation. Ideally the implementation should also
	 * use sparsity of the comparator if possible.
	 *
	 * @param <X> the values the given comparator can compare.
	 * @param comp the DerivableComparator that provides functionality to
	 * calculate local derivatives.
	 * @param keyword the keyword for which the given comparator was used in
	 * the Alignment.
	 *
	 * @return the derivatives for all parameters of the given comparator.
	 */
	public <X extends Value, Y> double[] calculateRawParameterDerivative(
		DerivableComparator<X, Y> comp, String keyword);

	/**
	 * This returns the derivatives for all parameters of the given comparator
	 * according to the implementation. Ideally the implementation should also
	 * use sparsity of the comparator if possible.
	 *
	 * @param <X> the values the given comparator can compare.
	 * @param <Y> the result format for derivatives of the given comparator.
	 * @param comp the DerivableComparator that provides functionality to
	 * calculate local derivatives.
	 * @param keyword the keyword for which the given comparator was used in
	 * the Alignment.
	 *
	 * @return the derivatives for all parameters of the given comparator in the
	 * result form given by the respective comparator.
	 */
	public <X extends Value, Y> Y calculateParameterDerivative(
		DerivableComparator<X, Y> comp, String keyword);

	/**
	 * Calculates the derivative for each keyword weight.
	 *
	 * @return the derivative for each keyword weight.
	 */
	public double[] calculateWeightDerivative();
}
