/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment;

import java.util.ArrayList;
import java.util.Random;
import java.util.TreeMap;

/**
 * This is a helper class, which actually extends the TreeMap to implement a
 * PriorityQueue that is able to poll the minimum as well as the maximum in
 * constanst time.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 * @param <K> The Key Class
 * @param <V> The Value Class
 */
public class IndexedStochasticPriorityQueue<K extends Comparable<K>, V>
        extends TreeMap<K, ArrayList<V>> {

    private final Random rand = new Random();

    public IndexedStochasticPriorityQueue() {
    }

    public void put(final K key, final V value) {
        ArrayList<V> list = super.get(key);
        if (list == null) {
            list = new ArrayList<V>();
            super.put(key, list);
        }
        list.add(value);
    }

    /**
     * Polls the value associated with the minimum key. If more than one element
     * is associated with the minimum key, it returns a random element from that
     * list.
     *
     * @return the value associated with the minimum key.
     */
    public V pollFirst() {
        if (isEmpty()) {
            return null;
        }
        final ArrayList<V> list = super.firstEntry().getValue();
        if (list.size() > 1) {
            final int pollIdx = rand.nextInt(list.size());
            return list.remove(pollIdx);
        } else {
            return super.pollFirstEntry().getValue().get(0);
        }
    }

    /**
     * Polls the value associated with the maximum key. If more than one element
     * is associated with the minimum key, it returns a random element from that
     * list.
     *
     * @return the value associated with the maximum key.
     */
    public V pollLast() {
        if (isEmpty()) {
            return null;
        }
        final ArrayList<V> list = super.lastEntry().getValue();
        if (list.size() > 1) {
            final int pollIdx = rand.nextInt(list.size());
            return list.remove(pollIdx);
        } else {
            return super.pollLastEntry().getValue().get(0);
        }
    }
}
