/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.OperationType;
import de.unibi.citec.tcs.alignment.sequence.Value;

/**
 * This is an interface for comparators that have derivable parameters.
 *
 * This interface has two purposes:
 * <ol><li>calculate the local derivative for a given parameter</li>
 * <li>identify this comparators parameters according to a common standard,
 * namely an indexing scheme, were each parameter gets assigned a unique integer
 * identifier between 0 and the number of parameters (exclusive). The
 * implementation has to ensure that this mapping is consistent. Outside users
 * may refer to the parameters in terms of their index.</li></ol>
 *
 * Note that the derivative of a parameter is always (!) a double value
 * according to the current schemes. Other parameters than doubles are not
 * supported in the current format.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 * @param <X> The value that can be compared by this class.
 * @param <Y> An intuitive format for the derivatives of all parameters of this
 * comparator.
 */
public interface DerivableComparator<X extends Value, Y> {

	/**
	 * This is called by an AlignmentDerivativeAlgorithm and should calculate
	 * the local derivative for only one comparison of the values a and b.
	 *
	 * Please note that a and b might be null if this alignment uses gaps or
	 * skips.
	 *
	 * @param paramIdx the index for this parameters.
	 * @param a the value that has been compared with b.
	 * @param b the value that has been compared with a.
	 * @param type the operation type that has actually been used in the
	 * comparison.
	 *
	 * @return the local derivative for a comparison of the values a and b.
	 */
	public double calculateLocalDerivative(int paramIdx, X a, X b, OperationType type);

	/**
	 * This method is supposed to construct the derivative for this comparator
	 * in an intuitive format given its vectorial form. The derviative for the
	 * parameter with index i is at position i in the array.
	 *
	 * @param derivatives the derivatives for all parameters of this comparator
	 * in its vectorial form.
	 *
	 * @return the derivatives for all parameters of this comparator in
	 * intuitive form.
	 */
	public Y transformToResult(double[] derivatives);

	/**
	 * This shall return the number of parameters of this comparator. The
	 * implementation of this interface has to ensure that parameters are
	 * uniquely identified with indices i between 0 and the number returned by
	 * this method (exclusive).
	 *
	 * @return the number of parameters of this comparator.
	 */
	public int getNumberOfParameters();

	/**
	 * Returns the class of the intuitive format of the derivatives for all
	 * parameters of this comparator. In the easiest case this is just a single
	 * double or a vector of doubles. It might also be a matrix or more complex
	 * datastructures.
	 *
	 * @return the class of the intuitive format of the derivatives for all
	 * parameters of this comparator
	 */
	public Class<Y> getResultClass();
}
