/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.OperationType;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import de.unibi.citec.tcs.alignment.sequence.ValueType;

/**
 * This comparator defines explicit replacement (and deletion/insertion) costs
 * for comparison of symbolic values.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class ReplacementComparator implements SkipComparator<SymbolicValue>,
	SparseDerivableComparator<SymbolicValue, double[][]> {

	private ReplacementCosts costMatrix;

	public ReplacementComparator() {
		this.costMatrix = null;
	}

	public ReplacementComparator(ReplacementCosts costMatrix) {
		this.costMatrix = costMatrix;
	}

	/**
	 * Sets the ReplacementCosts that shall be used within this Comparator.
	 * Please refer to the ReplacementCosts object for more information.
	 *
	 * @param costMatrix the ReplacementCosts that shall be used within this
	 * Comparator
	 */
	public void setCostMatrix(ReplacementCosts costMatrix) {
		this.costMatrix = costMatrix;
	}

	/**
	 * Returns the ReplacementCosts object that defines the explicit cost matrix
	 * for replacements between the symbols in the given alphabet. Please note
	 * that this is a Java object, not the actual underlying double matrix. If
	 * you want to retrieve that matrix you have to ask the returned object
	 * again for the CostMatrix.
	 *
	 * @return the ReplacementCosts object that defines the explicit cost matrix
	 * for replacements between the symbols in the given alphabet.
	 */
	public ReplacementCosts getCostMatrix() {
		return costMatrix;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double skipDelete(SymbolicValue a) {
		if (a == null) {
			return 0;
		}
		return costMatrix.getSkipDeletionCost(a.getSymbol());
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double skipInsert(SymbolicValue b) {
		if (b == null) {
			return 0;
		}
		return costMatrix.getSkipInsertionCost(b.getSymbol());
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double delete(SymbolicValue a) {
		if (a == null) {
			return 0;
		}
		return costMatrix.getDeletionCost(a.getSymbol());
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double insert(SymbolicValue b) {
		if (b == null) {
			return 0;
		}
		return costMatrix.getInsertionCost(b.getSymbol());
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double compare(SymbolicValue a, SymbolicValue b) {
		if (a == null && b == null) {
			return 0;
		}
		if (a == null && b != null) {
			return costMatrix.getInsertionCost(b.getSymbol());
		}
		if (a != null && b == null) {
			return costMatrix.getDeletionCost(a.getSymbol());
		}
		return costMatrix.getReplacementCost(a.getSymbol(), b.getSymbol());
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ValueType getType() {
		return ValueType.SYMBOLIC;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double calculateLocalDerivative(int paramIdx,
		SymbolicValue a, SymbolicValue b, OperationType type) {

		final int A = costMatrix.getAlphabet().size();
		final int APrime = costMatrix.getCostMatrix().length;

		final int iExpected = paramIdx / APrime;
		final int jExpected = paramIdx % APrime;
		if (iExpected >= A && jExpected >= A) {
			return 0;
		}

		final int iActual;
		if (a == null) {
			iActual = 0;
		} else {
			iActual = a.getAlphabet().getSymbolIndex(a.getSymbol());
		}
		final int jActual;
		if (b == null) {
			jActual = 0;
		} else {
			jActual = b.getAlphabet().getSymbolIndex(b.getSymbol());
		}

		switch (type) {
			case DELETION:
				if (iExpected != iActual || jExpected != A) {
					return 0;
				}
				break;
			case SKIPDELETION:
				if (iExpected != iActual || jExpected != A + 1) {
					return 0;
				}
				break;
			case INSERTION:
				if (iExpected != A || jExpected != jActual) {
					return 0;
				}
				break;
			case SKIPINSERTION:
				if (iExpected != A + 1 || jExpected != jActual) {
					return 0;
				}
				break;
			case REPLACEMENT:
				if (iExpected != iActual || jExpected != jActual) {
					return 0;
				}
				break;
		}
		return 1;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public SparseLocalDerivative calculateSparseLocalDerivative(
		SymbolicValue a, SymbolicValue b, OperationType type) {
		final SparseLocalDerivative.TrivialSparseLocalDerivative out
			= new SparseLocalDerivative.TrivialSparseLocalDerivative();

		final int A = costMatrix.getAlphabet().size();
		final int APrime = costMatrix.getCostMatrix().length;

		int i;
		if (a == null) {
			i = 0;
		} else {
			i = a.getAlphabet().getSymbolIndex(a.getSymbol());
		}
		int j;
		if (b == null) {
			j = 0;
		} else {
			j = b.getAlphabet().getSymbolIndex(b.getSymbol());
		}

		switch (type) {
			case DELETION:
				j = A;
				break;
			case SKIPDELETION:
				j = A + 1;
				break;
			case INSERTION:
				i = A;
				break;
			case SKIPINSERTION:
				i = A + 1;
				break;
		}
		out.add(i * APrime + j, 1);
		return out;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int hashCode() {
		int hash = 5;
		hash = 97 * hash + (this.costMatrix != null ? this.costMatrix.hashCode() : 0);
		return hash;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final ReplacementComparator other = (ReplacementComparator) obj;
		if (this.costMatrix != other.costMatrix
			&& (this.costMatrix == null || !this.costMatrix.equals(other.costMatrix))) {
			return false;
		}
		return true;
	}

	@Override
	public Class<double[][]> getResultClass() {
		return double[][].class;
	}

	@Override
	public double[][] transformToResult(double[] derivatives) {
		final int APrime = costMatrix.getCostMatrix().length;
		final double[][] out = new double[APrime][APrime];
		for (int i = 0; i < APrime; i++) {
			for (int j = 0; j < APrime; j++) {
				out[i][j] = derivatives[i * APrime + j];
			}
		}

		return out;
	}

	@Override
	public int getNumberOfParameters() {
		final int APrime = costMatrix.getCostMatrix().length;
		return APrime * APrime;
	}
}
