/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.comparators.SparseLocalDerivative.TrivialSparseLocalDerivative;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import de.unibi.citec.tcs.alignment.sequence.ValueType;

/**
 * This is a Comparator for SymbolicValues that just specifies a score for:
 * <ol>
 * <li>the deletion of a symbol (1 per default)</li>
 * <li>the insertion of a symbol (1 per default)</li>
 * <li>the skip deletion of a symbol (0.5 per default)</li>
 * <li>the skip insertion of a symbol (0.5 per default)</li>
 * <li>a match between the same symbols (0 per default)</li>
 * <li>a mismatch between different symbols (1 per default)</li>
 * </ol>
 *
 * Please note that this is just a special case of the much more general
 * ReplacementComparator.
 *
 * For generality we use the SkipExtendedComparator as super class here.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class TrivialEditComparator extends SkipExtendedComparator<SymbolicValue> {

	private double matchCost = 0;
	private double mismatchCost = 1;

	public TrivialEditComparator() {
	}

	public double getMatchCost() {
		return matchCost;
	}

	public void setMatchCost(double matchCost) {
		this.matchCost = matchCost;
	}

	public double getMismatchCost() {
		return mismatchCost;
	}

	public void setMismatchCost(double mismatchCost) {
		this.mismatchCost = mismatchCost;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ValueType getType() {
		return ValueType.SYMBOLIC;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double compare(SymbolicValue a, SymbolicValue b) {
		if (a.equals(b)) {
			return matchCost;
		} else {
			return mismatchCost;
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double comparisonDerivative(int paramIdx, SymbolicValue a, SymbolicValue b) {
		switch (paramIdx) {
			case 4:
				//match
				if (a.equals(b)) {
					return 1;
				} else {
					return 0;
				}
			case 5:
				//mismatch
				if (a.equals(b)) {
					return 0;
				} else {
					return 1;
				}
			default:
				return 0;
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public SparseLocalDerivative sparseComparisonDerivative(SymbolicValue a, SymbolicValue b) {
		final TrivialSparseLocalDerivative out = new TrivialSparseLocalDerivative();
		if (a.equals(b)) {
			out.add(4, 1);
		} else {
			out.add(5, 1);
		}
		return out;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int getNumberOfComparisonParameters() {
		return 2;
	}

}
