/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.visualization;
//package de.unibi.citec.fit.dfgvis;

import static de.unibi.citec.tcs.alignment.visualization.Visualization.ERROR_FILE_PERMISSION;
//import static de.unibi.citec.fit.dfgvis.Visualization.ERROR_FILE_PERMISSION;
import de.unibi.citec.tcs.alignment.sequence.Node;
import java.io.File;
import java.util.HashMap;
import java.util.Map;

/**
 *
 * @author ckrueger
 */
public abstract class AbstractHTMLTransformerFromPath extends AbstractSpecialHTMLTransformer{
    
    /**
     * Maps filenames to their file representation.
     */
    private Map<String, File> files = new HashMap<String, File>();
    
    /**
     * Attention: when using this constructor, you have to call 
     * setFiles() manually.
     * @param keyword the keyword, as it should appear in the table.
     */
    public AbstractHTMLTransformerFromPath(final String keyword) {
        super(keyword);
    }
    
    /**
     * Initializes given files or folders. 
     * @param keyword the keyword, as it should appear in the table.
     * @param files_or_folders Paths to source code files or 
     *      a directory/directories containing such files.
     */
    public AbstractHTMLTransformerFromPath(final String keyword, 
            final String[] files_or_folders) {
        
        super(keyword);
        for (String str : files_or_folders) {
            addToFiles(str);
        }
    }

    /**
     * Initializes given folder. 
     * @param keyword the keyword, as it should appear in the table.
     * @param folder Paths a directory containing such source code files.
     */
    public AbstractHTMLTransformerFromPath(final String keyword, 
            final String folder) {
        
        super(keyword);
        addToFiles(folder);
    }

    /**
     * Initializes given files or folders. 
     * @param keyword the keyword, as it should appear in the table.
     * @param files_or_folders Source code files or 
     *      a directory/directories containing such files.
     */
    public AbstractHTMLTransformerFromPath(final String keyword, 
            final File[] files_or_folders) {
        
        super(keyword);        
        for (File file : files_or_folders) {
            addToFiles(file);
        }
    }

    /**
     * Initializes given folder. 
     * @param keyword the keyword, as it should appear in the table.
     * @param folder A directory containing such source code files.
     */
    public AbstractHTMLTransformerFromPath(final String keyword, 
            final File folder) {
        
        super(keyword);
        addToFiles(folder);
    }
        
    /**
     * Adds the file specified by the given path to the files map.
     * If the given path points to a directory, this directory is stored with 
     * the name 'dir[:I:]' where [:I:] refers to the number of already added 
     * directories.
     * @param path Path to a source code file or a directory containing such files.
     */
    private void addToFiles(final String path) {
        if (path == null || path.length() == 0) {
            throw new RuntimeException("Given path is null or empty."); 
        } 
           
        final File file = new File(path);
        addToFiles(file);
    }
    
    /**
     * Adds the given file to the files map.
     * If the given file points to a directory, this directory is stored with 
     * the name 'dir[:I:]' where [:I:] refers to the number of already added 
     * directories.
     * @param file A source code file or a directory containing such files.
     */
    private void addToFiles(final File file) {
        final String path = file.getAbsolutePath();
        if (!(file.exists() && file.canRead())) 
            throw new RuntimeException(ERROR_FILE_PERMISSION + ": " + path); 
        
        // path is a directory
        if (file.isDirectory()) {
            if (!file.canExecute()) {
                throw new RuntimeException(ERROR_FILE_PERMISSION + ": " + path); 
            }

            Integer i = 0; 
            while (files.containsKey("dir" + i.toString())) {
                i++;
            }
            files.put("dir" + i.toString(), file);

        //path is a file
        } else {
            final String filename = path.substring(
                    path.lastIndexOf(File.separator));
            if (files.containsKey(filename) && !file.equals(files.get(filename))) {
                throw new RuntimeException("For file " + filename 
                        + " are two paths given:\n\t" + path
                        + "\n\t" + files.get(filename).getAbsolutePath());
            }
            files.put(filename, file);
        }
    }
    
    /**
     * Sets the files map directly.
     * @param files Paths to files or folder[s] containing files.
     */
    public void setFiles(final Map<String, File> files) {
        this.files = files;
    }
    
    /**
     * Returns the files map.
     * @return The files map.
     */
    public Map<String, File> getFiles() {
        return this.files;
    }

    @Override
    abstract public String getHTMLforNode(Node node);
    
}
