/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.visualization;
//package de.unibi.citec.fit.dfgvis;

import de.unibi.citec.tcs.alignment.sequence.Sequence;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * ExternalContent holds a bunch of selfdefined keywords and their corresponding
 * content for one sequence, thereby for example one solution graph.
 * External here stands for, not computed by the alignment algorithm.
 * The keyword can be a self conceived key or one defined in a 
 * sequence/alignment/operation. If the keyword is already defined the 
 * external content overwrites the defined one (if the keyword is also 
 * transfered by usedKeys to the output method. 
 * @author ckrueger
 */
public class ExternalContent {
    /**
     * Map of Keywords and their corresponding content for a sequence of
     * an alignment. The size of the corresponding content equals to 
     * the number of nodes in the sequence.
     */
    private final Map<String, ArrayList<String> > content;
    
    public ExternalContent() {
        content = new HashMap<String, ArrayList<String> >();
    }
    
    /**
     * Returns the external content for a specific keyword.
     * @param key 
     * @return null if the key is unknown.
     */
    public ArrayList<String> getContent(String key) {
        if (contains(key)) {
            return content.get(key);
        }
        return null;
    }
    
    /**
     * Returns the external content for a specific keyword and index.
     * @param key 
     * @param idx The index corresponds to the entry in sequence with the index idx.
     * @return null if the key is unknown.
     */
    public String getContent(String key, int idx) {
        if (contains(key)) {
            if (content.get(key).size() > idx) return content.get(key).get(idx);
        }
        return null;
    }
    
    /**
     * Adds content for the key to this ExternalContent.
     * @param hTransForm Interface instance that transforms the nodes of the 
     *      sequence to the desired format.
     * @param seq Sequence whose nodes should be transformed.
     */
    public void add(final SpecialHTMLTransformer hTransForm, Sequence seq) {
        ArrayList<String> newContent = new ArrayList<String>();
        
        for (final de.unibi.citec.tcs.alignment.sequence.Node node : seq.getNodes()) {
            newContent.add(hTransForm.getHTMLforNode(node));
        }
        this.content.put(hTransForm.getKeyword(), newContent);
    }    
    
    public boolean contains(String key) {
        return (content.containsKey(key));
    }
}
