/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import junit.framework.TestCase;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class AlignmentPathTest extends TestCase {

	public AlignmentPathTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of calculateWeightDerivative method, of class AlignmentPath.
	 */
	public void testCalculateWeightDerivative() {
		/*
		 * we use a toy example here, where one feature does not differ at all
		 * while another one does, namely a java program snippet with variable
		 * declarations with different classes.
		 */
		//specify the keywords.
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[2];
		//first the node type.
		final Alphabet nodeTypes = new Alphabet(new String[]{"double", "int"});
		keywordSpecs[0] = new SymbolicKeywordSpecification(nodeTypes, "class");
		//then the name
		keywordSpecs[1] = new StringKeywordSpecification("name");
		//derive the node specification from that.
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		//than create the sequences (one double and one int based)
		final Sequence a = new Sequence(nodeSpec);
		Node node = new Node(a);
		node.setValue("class", new SymbolicValue(nodeTypes, "int"));
		node.setValue("name", new StringValue("i"));
		a.getNodes().add(node);
		node = new Node(a);
		node.setValue("class", new SymbolicValue(nodeTypes, "int"));
		node.setValue("name", new StringValue("j"));
		a.getNodes().add(node);

		final Sequence b = new Sequence(nodeSpec);
		node = new Node(b);
		node.setValue("class", new SymbolicValue(nodeTypes, "double"));
		node.setValue("name", new StringValue("i"));
		b.getNodes().add(node);
		node = new Node(b);
		node.setValue("class", new SymbolicValue(nodeTypes, "double"));
		node.setValue("name", new StringValue("j"));
		b.getNodes().add(node);

		//calculate the alignment
		final AlignmentSpecification alignSpec = new AlignmentSpecification(nodeSpec);
		final StrictAlignmentFullAlgorithm strict = new StrictAlignmentFullAlgorithm(alignSpec);
		final AlignmentPath instance = strict.calculateAlignment(a, b);

		/*
		 * check the derivative.
		 *
		 * Two replacements should have happened. In both cases we had a
		 * mismatch of class and a perfect match for the name. Thus the weight
		 * derivative for the first keyword (class) has to be 2 and for the
		 * second keyword (name) it has to be 0.
		 */
		final double[] actual = instance.calculateWeightDerivative();

		assertEquals(2, actual.length);
		assertEquals(2, actual[0], 0);
		assertEquals(0, actual[1], 0);
	}

	/**
	 * Test of calculateDerivative method, of class AlignmentPath.
	 */
	public void testCalculateDerivative() {
		/*
		 * We use a toy example here with an alignment of the strings
		 *
		 * ab
		 *
		 * and
		 *
		 * aab
		 *
		 * where a->a and b->b replacement are used as well as an a insertion,
		 * but nothing else.
		 */
		final Sequence a = new Sequence("ab");
		final Sequence b = new Sequence("aab");

		//use Strict alignment
		final StrictAlignmentFullAlgorithm algo = new StrictAlignmentFullAlgorithm(
				new AlignmentSpecification(a));
		final AlignmentPath instance = algo.calculateAlignment(a, b);
		//calculate derivative
		final ReplacementComparator comparator
				= (ReplacementComparator) instance.getSpecificaton().getComparator(0);
		final double[][] derivative = instance.calculateParameterDerivative(
				comparator, Sequence.DEFAULTKEYWORD);

		//replacement derivative should be 1.
		assertEquals(1, derivative[0][0], 0);
		assertEquals(1, derivative[1][1], 0);
		//deletion deriative should be 0
		assertEquals(0, derivative[0][2], 0);
		assertEquals(0, derivative[1][2], 0);
		//we insert an a, but not a b
		assertEquals(1, derivative[2][0], 0);
		assertEquals(0, derivative[2][1], 0);

		System.out.println(instance.toString());
	}
}
