/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.wrappers.StringEditDistance;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.Random;
import junit.framework.TestCase;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class ParallelProcessingEngineTest extends TestCase {

	public ParallelProcessingEngineTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of calculate method, of class ParallelProcessingEngine.
	 */
	public void testCalculate() {

		//set up sequences.
		final Sequence[] sequences = new Sequence[3];
		sequences[0] = new Sequence("ab");
		sequences[1] = new Sequence("aab");
		sequences[2] = new Sequence("abb");

		//set up algorithm.
		final StrictAlignmentScoreAlgorithm algo
				= new StrictAlignmentScoreAlgorithm(new AlignmentSpecification(sequences[0]));

		final int threadNum = 8;

		//first check with default options if it works at all.
		final ParallelProcessingEngine<Double> instance
				= new ParallelProcessingEngine<Double>(algo, sequences);
		instance.setVerbose(false);
		assertSame(algo, instance.getAlgorithm());
		instance.setNumberOfThreads(threadNum);
		assertEquals(threadNum, instance.getNumberOfThreads());

		instance.calculate();
		assertTrue(instance.isFinished());
		Double[][] resultMatrix = instance.getResultMatrix();
		double[][] primitiveResultMatrix = instance.getScoreMatrix();
		assertEquals(sequences.length, resultMatrix.length);
		for (int i = 0; i < sequences.length; i++) {
			assertEquals(sequences.length, resultMatrix[i].length);
			for (int j = 0; j < sequences.length; j++) {
				assertNotNull(resultMatrix[i][j]);
				if (i == j) {
					assertEquals(0, resultMatrix[i][j], 0);
				} else {
					assertTrue(resultMatrix[i][j] > 0);
				}
				assertEquals(resultMatrix[i][j], primitiveResultMatrix[i][j], 0);
			}
		}
		//exclude some calculations.

		final boolean[][] doCalculation
				= new boolean[sequences.length][sequences.length];
		doCalculation[0][0] = true;

		instance.setDoCalculation(doCalculation);
		instance.calculate();
		resultMatrix = instance.getResultMatrix();
		primitiveResultMatrix = instance.getScoreMatrix();
		assertNotNull(resultMatrix);
		assertEquals(sequences.length, resultMatrix.length);
		assertNotNull(resultMatrix[0][0]);
		assertEquals(0, resultMatrix[0][0], 0);
		assertEquals(resultMatrix[0][0], primitiveResultMatrix[0][0], 0);
		for (int i = 0; i < sequences.length; i++) {
			assertEquals(sequences.length, resultMatrix[i].length);
			for (int j = 0; j < sequences.length; j++) {
				if (i == 0 && j == 0) {
					continue;
				}
				assertNull(resultMatrix[i][j]);
				assertTrue(Double.isNaN(primitiveResultMatrix[i][j]));
			}
		}
	}

	/**
	 * Test of calculate method, of class ParallelProcessingEngine.
	 */
	public void testCalculationSpeedUp() {

		//generate longer test sequences.
		final int seqNum = 50;
		final int seqLength = 70;
		final Sequence[] sequences = new Sequence[seqNum];
		final Random rand = new Random();
		for (int i = 0; i < seqNum; i++) {
			final char[] seqIndices = new char[seqLength];
			for (int j = 0; j < seqLength; j++) {
				seqIndices[j] = (char) (65 + rand.nextInt(2));
			}
			sequences[i] = new Sequence(new String(seqIndices));
		}

		//set up algorithm.
		final StrictAlignmentScoreAlgorithm algo
				= new StrictAlignmentScoreAlgorithm(new AlignmentSpecification(sequences[0]));

		final int numThreads = 32;
		final ParallelProcessingEngine<Double> instance
				= new ParallelProcessingEngine<Double>(algo, sequences);
		instance.setNumberOfThreads(numThreads);
		instance.setVerbose(false);

		//do a sequential computation.
		final double[][] benchmarkMat = new double[seqNum][seqNum];
		long start = System.nanoTime();
		for (int i = 0; i < sequences.length; i++) {
			for (int j = 0; j < sequences.length; j++) {
				benchmarkMat[i][j] = algo.calculateAlignment(sequences[i],
						sequences[j]);
			}
		}
		long end = System.nanoTime();
		long sequentialTime = end - start;
		//do a parallel computation
		start = System.nanoTime();
		instance.calculate();
		end = System.nanoTime();
		long parallelTime = end - start;
		//do a parallel computation with symmetry assumption.
		instance.setSymmetric(true);
		start = System.nanoTime();
		instance.calculate();
		end = System.nanoTime();
		long symmetricTime = end - start;

		System.out.println("sequential:" + sequentialTime / 1000000 + " ms");
		System.out.println("parallel:" + parallelTime / 1000000 + " ms");
		System.out.println("symmetric:" + symmetricTime / 1000000 + " ms");
		//we expect at least a speedup for parallel computation.
		assertTrue(parallelTime < sequentialTime);
		//and a 30% speedup for symmetric computation.
		assertTrue(symmetricTime < parallelTime * 0.7);
		//check the results.
		final Double[][] symmMatrix = instance.getResultMatrix();
		assertEquals(seqNum, benchmarkMat.length);
		assertEquals(seqNum, symmMatrix.length);
		for (int i = 0; i < seqNum; i++) {
			assertEquals(seqNum, benchmarkMat[i].length);
			assertEquals(seqNum, symmMatrix[i].length);
			for (int j = 0; j < seqNum; j++) {
				assertEquals(symmMatrix[i][j], symmMatrix[j][i], 0);
				assertEquals(benchmarkMat[i][j], symmMatrix[i][j], 0);
			}
		}
	}

	/**
	 * Test on how the ParallelProcessing reacts to exceptions.
	 */
	public void testExceptionHandling() throws UnsupportedEncodingException {
		//create two sequences with non-matching NodeSpecifications.
		final Sequence[] dataSpace = new Sequence[2];
		dataSpace[0] = StringEditDistance.toSequences(new String[]{"aaaa"})[0];
		dataSpace[1] = StringEditDistance.toSequences(new String[]{"abaa"})[0];
		//construct alignment specification for them.
		final AlignmentSpecification alignSpec = new AlignmentSpecification(
				dataSpace[0].getNodeSpecificiation());
		//construct algorithm.
		final StrictAlignmentScoreAlgorithm algo
				= new StrictAlignmentScoreAlgorithm(alignSpec);
		//if we try to align now, this should lead to an exception.
		boolean exception = false;
		try {
			algo.calculateAlignment(dataSpace[0], dataSpace[1]);
		} catch (UnsupportedOperationException ex) {
			assertEquals("The node specifications of both input sequences to not match!", ex.getMessage());
			exception = true;
		}
		assertTrue("Expected an exception!", exception);

		//if we try to use this in a ParallelProcessingEngine, the same exception should come up.
		final ParallelProcessingEngine engine = new ParallelProcessingEngine(algo, dataSpace);
		exception = false;
		try {
			engine.calculate();
		} catch (UnsupportedOperationException ex) {
			assertTrue(
					"The node specifications of both input sequences to not match!".equals(ex.getMessage())
					|| "The first input sequence has an unexpected node specification!".equals(ex.getMessage()));
			exception = true;
		}
	}

}
