/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author bpaassen
 */
public class SoftAlignmentSamplingAlgorithmTest {

    public SoftAlignmentSamplingAlgorithmTest() {
    }

    @BeforeClass
    public static void setUpClass() {
    }

    @AfterClass
    public static void tearDownClass() {
    }

    @Before
    public void setUp() {
    }

    @After
    public void tearDown() {
    }

    /**
     * Test of calculateAlignment method, of class StrictKPathAlgorithm.
     */
    @org.junit.Test
    public void testCalculateAlignment() {
        /*
         * we set up a simple input: An alignment of a and a. The optimal
         * alignment obviously is a simple replacement. However two other
         * alignments are possible (even though they are suboptimal): First
         * deleting a and then inserting a or first inserting a and than
         * deleting a.
         */
        final Sequence a = new Sequence("a");
        final Sequence b = new Sequence("a");
        final SoftAlignmentSamplingAlgorithm instance = new SoftAlignmentSamplingAlgorithm(
                new AlignmentSpecification(a));
        instance.setBeta(15);
        instance.setPathLimit(1);
        PathList pathList = instance.calculateAlignment(a, b);
        assertEquals(1, pathList.size());
        /*
         * For a high beta we can be sure that the optimal path is chosen,
         * namely an a to a replacement.
         */
        final AlignmentPath path = pathList.get(0);
        assertEquals(1, path.getOperations().size());
        assertEquals(0, path.getScore(), 0);
        assertEquals(OperationType.REPLACEMENT, path.getOperations().get(0).getType());
        /*
         * For a slightly lower beta the average score of many samples should
         * still be close to optimal.
         */
        instance.setBeta(4.6);
        instance.setPathLimit(1000);
        pathList = instance.calculateAlignment(a, b);
        assertEquals(1000, pathList.size());
        assertEquals(0, pathList.getAverageScore(), 0.1);
    }

    /**
     * Test of calculateAlignment method, of class StrictKPathAlgorithm.
     */
    @org.junit.Test
    public void testAlignmentPerformance() {

        //align a...a and a...b
        final int length = 1000;

        final StringBuilder aBuilder = new StringBuilder();
        final StringBuilder bBuilder = new StringBuilder();

        aBuilder.append('b');
        bBuilder.append('b');
        for (int i = 1; i < length - 1; i++) {
            aBuilder.append('a');
            bBuilder.append('a');
        }
        aBuilder.append('a');
        bBuilder.append('b');

        final Sequence a = new Sequence(aBuilder.toString());
        final Sequence b = new Sequence(bBuilder.toString());

        //initialize actual instance.
        final SoftAlignmentSamplingAlgorithm instance
                = new SoftAlignmentSamplingAlgorithm(new AlignmentSpecification(a));
        instance.setBeta(15);

        final long start = System.nanoTime();
        final PathList pathList = instance.calculateAlignment(a, b);
        assertEquals(SoftAlignmentSamplingAlgorithm.DEFAULTPATHLIMIT, pathList.size());
        final double time = (double) (System.nanoTime() - start) / 1E9;

        //check the derivative calculation
        final ReplacementComparator repComp
                = (ReplacementComparator) instance.getSpecification().getComparator(0);
        final double[][] derivative = pathList.calculateParameterDerivative(
				repComp, Sequence.DEFAULTKEYWORD);
        final double tolerance = 1;
        //the a->a derivative should be highest by far and close to the
        //length.
        assertEquals(length - 2, derivative[0][0], tolerance);
        //a-> b and b -> b derivative should be about 1.
        assertEquals(1, derivative[0][1], tolerance);
        assertEquals(1, derivative[1][1], tolerance);
        //b-> a derivative should be 0.
        assertEquals(0, derivative[1][0], tolerance);
        //gap derivatives should be close to 0.
        for (int s = 0; s < 2; s++) {
            assertEquals(0, derivative[s][2], tolerance);
            assertEquals(0, derivative[2][s], tolerance);
        }
    }
}
