/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.comparators.ReplacementCosts;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import static junit.framework.Assert.assertEquals;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class SoftLocalPathModelTest {

	public SoftLocalPathModelTest() {
	}

	@BeforeClass
	public static void setUpClass() {
	}

	@AfterClass
	public static void tearDownClass() {
	}

	@Before
	public void setUp() {
	}

	@After
	public void tearDown() {
	}

	/**
	 * Test of calculateParameterDerivative method, of class SoftLocalPathModel.
	 */
	@Test
	public void testCalculateParameterDerivative() {
		//set up data space.
		final Alphabet alpha = new Alphabet("a|b|c|d|e|y|z");
		final int ALPHASIZE = alpha.size() + 2;
		final NodeSpecification nodeSpec = new NodeSpecification(
				new KeywordSpecification[]{
					new SymbolicKeywordSpecification(alpha, "content")
				});
		final ReplacementComparator comp = new ReplacementComparator(
				new ReplacementCosts(alpha, true, true));
		AlignmentSpecification alignSpec = new AlignmentSpecification(nodeSpec);
		alignSpec.setComparator(0, comp);
		comp.getCostMatrix().setSkipDeletionCost("a", 0.1);
		comp.getCostMatrix().setSkipDeletionCost("b", 0.1);
		comp.getCostMatrix().setSkipInsertionCost("a", 0.1);
		comp.getCostMatrix().setSkipInsertionCost("b", 0.1);
		final Sequence[] seqs = new Sequence[2];
		seqs[0] = new Sequence("aaaa", alpha);
		seqs[1] = new Sequence("bbbb", alpha);
		final SoftLocalAlignmentFullAlgorithm algo = new SoftLocalAlignmentFullAlgorithm(alignSpec);
		algo.setBeta(10);

		final Sequence empty = new Sequence("", alpha);

		//align the empty sequence with itself.
		SoftLocalPathModel instance = algo.calculateAlignment(empty, empty);
		double[][] deriv = instance.calculateParameterDerivative(comp, "content");
		//we expect no operations.
		assertEquals(ALPHASIZE, deriv.length);
		for (int i = 0; i < ALPHASIZE; i++) {
			assertEquals(ALPHASIZE, deriv[i].length);
			for (int j = 0; j < ALPHASIZE; j++) {
				assertEquals(0, deriv[i][j], 0);
			}
		}

		//align an empty string with a non-empty string.
		instance = algo.calculateAlignment(empty, seqs[0]);
		deriv = instance.calculateParameterDerivative(comp, "content");
		//we expect four skip-insertions and nothing else.
		assertEquals(ALPHASIZE, deriv.length);
		for (int i = 0; i < ALPHASIZE; i++) {
			assertEquals(ALPHASIZE, deriv[i].length);
			for (int j = 0; j < ALPHASIZE; j++) {
				if (i == ALPHASIZE - 1 && j == 0) {
					assertEquals(4, deriv[i][j], 0);
				} else {
					assertEquals(0, deriv[i][j], 0);
				}
			}
		}
		//that should be symmetric.
		instance = algo.calculateAlignment(seqs[0], empty);
		deriv = instance.calculateParameterDerivative(comp, "content");
		assertEquals(ALPHASIZE, deriv.length);
		for (int i = 0; i < ALPHASIZE; i++) {
			assertEquals(ALPHASIZE, deriv[i].length);
			for (int j = 0; j < ALPHASIZE; j++) {
				if (i == 0 && j == ALPHASIZE - 1) {
					assertEquals(4, deriv[i][j], 0);
				} else {
					assertEquals(0, deriv[i][j], 0);
				}
			}
		}
		/*
		 * If we align the strings "aaaa" and "bbbb" we expect only skips.
		 */
		instance = algo.calculateAlignment(seqs[0], seqs[1]);
		deriv = instance.calculateParameterDerivative(comp, "content");
		assertEquals(ALPHASIZE, deriv.length);
		for (int i = 0; i < ALPHASIZE; i++) {
			assertEquals(ALPHASIZE, deriv[i].length);
			for (int j = 0; j < ALPHASIZE; j++) {
				if ((i == 0 && j == ALPHASIZE - 1) || (i == ALPHASIZE - 1 && j == 1)) {
					assertEquals(4, deriv[i][j], 0);
				} else {
					assertEquals(0, deriv[i][j], 0);
				}
			}
		}
		/*
		 * If we align the first sequence with itself we expect only
		 * replacements.
		 */
		algo.setBeta(50);
		instance = algo.calculateAlignment(seqs[0], seqs[0]);
		deriv = instance.calculateParameterDerivative(comp, "content");
		assertEquals(ALPHASIZE, deriv.length);
		for (int i = 0; i < ALPHASIZE; i++) {
			assertEquals(ALPHASIZE, deriv[i].length);
			for (int j = 0; j < ALPHASIZE; j++) {
				if (i == 0 && j == 0) {
					assertEquals(4, deriv[i][j], 0);
				} else {
					assertEquals(0, deriv[i][j], 0);
				}
			}
		}
		/*
		 * The same should hold for an alignment of the second sequence with
		 * itself.
		 */
		instance = algo.calculateAlignment(seqs[1], seqs[1]);
		deriv = instance.calculateParameterDerivative(comp, "content");
		assertEquals(ALPHASIZE, deriv.length);
		for (int i = 0; i < ALPHASIZE; i++) {
			assertEquals(ALPHASIZE, deriv[i].length);
			for (int j = 0; j < ALPHASIZE; j++) {
				if (i == 1 && j == 1) {
					assertEquals(4, deriv[i][j], 0);
				} else {
					assertEquals(0, deriv[i][j], 0);
				}
			}
		}
		/*
		 * Now check a non-trivial alignment where middle skips are necessary.
		 * There are two optimal paths:
		 *
		 * c vs. _ = 0.5
		 * _ vs. b = 0.1
		 * a vs. a = 0.0
		 * z vs. _ = 0.5
		 * z vs. _ = 0.5
		 * z vs. _ = 0.5
		 * z vs. _ = 0.5
		 * a vs. a = 0.0
		 * d vs. e = 1.0
		 *
		 * and
		 *
		 * c vs. _ = 0.5
		 * _ vs. b = 0.1
		 * a vs. a = 0.0
		 * z vs. _ = 0.5
		 * z vs. _ = 0.5
		 * z vs. _ = 0.5
		 * z vs. _ = 0.5
		 * a vs. a = 0.0
		 * d vs. _ = 0.5
		 * _ vs. e = 0.5
		 *
		 * Thus we expect:
		 *
		 * deriv("c","_") = 1
		 * deriv("_","b") = 1
		 * deriv("a","a") = 2
		 * deriv("z","_") = 4
		 * deriv("d","e") = 0.5
		 * deriv("d","_") = 0.5
		 * deriv("_","e") = 0.5
		 *
		 * and 0 for everything else.
		 */
		Sequence left = new Sequence("cazzzzad", alpha);
		Sequence right = new Sequence("baae", alpha);

		instance = algo.calculateAlignment(left, right);
		deriv = instance.calculateParameterDerivative(comp, "content");
		assertEquals(ALPHASIZE, deriv.length);
		for (int i = 0; i < ALPHASIZE; i++) {
			assertEquals(ALPHASIZE, deriv[i].length);
			for (int j = 0; j < ALPHASIZE; j++) {
				if (i == 0 && j == 0) {
					assertEquals(2, deriv[i][j], 0);
				} else if (i == 2 && j == ALPHASIZE - 1) {
					assertEquals(1, deriv[i][j], 0);
				} else if (i == 3 && j == 4) {
					assertEquals(0.5, deriv[i][j], 0);
				} else if (i == 3 && j == ALPHASIZE - 1) {
					assertEquals(0.5, deriv[i][j], 0);
				} else if (i == 6 && j == ALPHASIZE - 1) {
					assertEquals(4, deriv[i][j], 0);
				} else if (i == ALPHASIZE - 1 && j == 1) {
					assertEquals(1, deriv[i][j], 0);
				} else if (i == ALPHASIZE - 1 && j == 4) {
					assertEquals(0.5, deriv[i][j], 0);
				} else {
					assertEquals(0, deriv[i][j], 0);
				}
			}
		}

		/*
		 * Finally check an example where a deletion has to be used.
		 */
		for (final String sym : alpha.getSymbols()) {
			comp.getCostMatrix().setSkipDeletionCost(sym, 0.2);
			comp.getCostMatrix().setSkipInsertionCost(sym, 0.2);
		}

		final Sequence left2 = new Sequence("aaaaaaaazaaaaaaaa", alpha);
		final Sequence right2 = new Sequence("aaaaaaaayyaaaaaaaaaa", alpha);

		instance = algo.calculateAlignment(left2, right2);
		deriv = instance.calculateParameterDerivative(comp, "content");
		assertEquals(ALPHASIZE, deriv.length);
		for (int i = 0; i < ALPHASIZE; i++) {
			assertEquals(ALPHASIZE, deriv[i].length);
			for (int j = 0; j < ALPHASIZE; j++) {
				if (i == 0 && j == 0) {
					assertEquals(16, deriv[i][j], 1E-6);
				} else if (i == 6 && j == ALPHASIZE - 2) {
					assertEquals(1, deriv[i][j], 1E-6);
				} else if (i == ALPHASIZE - 1 && j == 0) {
					assertEquals(2, deriv[i][j], 1E-6);
				} else if (i == ALPHASIZE - 1 && j == 5) {
					assertEquals(2, deriv[i][j], 1E-6);
				} else {
					assertEquals(0, deriv[i][j], 0);
				}
			}
		}
	}

	/**
	 * Test of calculateWeightDerivative method, of class SoftLocalPathModel.
	 */
	@Test
	public void testCalculateWeightDerivative() {
		/*
		 * we use a toy example here, where one feature does not differ at all
		 * while another one does, namely a java program snippet with variable
		 * declarations with different classes.
		 */
		//specify the keywords.
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[2];
		//first the node type.
		final Alphabet nodeTypes = new Alphabet(new String[]{"double", "int"});
		keywordSpecs[0] = new SymbolicKeywordSpecification(nodeTypes, "class");
		//then the name
		keywordSpecs[1] = new StringKeywordSpecification("name");
		//derive the node specification from that.
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		//than create the sequences (one double and one int based)
		final Sequence a = new Sequence(nodeSpec);
		Node node = new Node(a);
		node.setValue("class", new SymbolicValue(nodeTypes, "int"));
		node.setValue("name", new StringValue("i"));
		a.getNodes().add(node);
		node = new Node(a);
		node.setValue("class", new SymbolicValue(nodeTypes, "int"));
		node.setValue("name", new StringValue("j"));
		a.getNodes().add(node);

		final Sequence b = new Sequence(nodeSpec);
		node = new Node(b);
		node.setValue("class", new SymbolicValue(nodeTypes, "double"));
		node.setValue("name", new StringValue("i"));
		b.getNodes().add(node);
		node = new Node(b);
		node.setValue("class", new SymbolicValue(nodeTypes, "double"));
		node.setValue("name", new StringValue("j"));
		b.getNodes().add(node);

		//calculate a soft alignment first.
		final AlignmentSpecification alignSpec = new AlignmentSpecification(nodeSpec);

		final SoftLocalAlignmentFullAlgorithm algo = new SoftLocalAlignmentFullAlgorithm(alignSpec);
		algo.setBeta(15);
		final SoftLocalPathModel instance = algo.calculateAlignment(a, b);

		//and calculate a strict alignment for comparison.
		final StrictLocalAlignmentFullAlgorithm strict = new StrictLocalAlignmentFullAlgorithm(alignSpec);
		final AlignmentPath alignment = strict.calculateAlignment(a, b);
		System.out.println(alignment.toString());

		//check the derivative.
		final double[] expected = alignment.calculateWeightDerivative();
		final double[] actual = instance.calculateWeightDerivative();

		assertEquals(2, expected.length);
		assertEquals(expected.length, actual.length);

		for (int k = 0; k < expected.length; k++) {
			assertEquals(expected[k], actual[k], 0);
		}
	}
}
