/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.wrappers;

import de.unibi.citec.tcs.alignment.AlignmentPath;
import de.unibi.citec.tcs.alignment.AlignmentSpecification;
import de.unibi.citec.tcs.alignment.Operation;
import de.unibi.citec.tcs.alignment.StrictAlignmentFullAlgorithm;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.Value;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;
import java.util.Arrays;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class VectorialSequencesTest {

	public VectorialSequencesTest() {
	}

	@BeforeClass
	public static void setUpClass() {
	}

	@AfterClass
	public static void tearDownClass() {
	}

	@Before
	public void setUp() {
	}

	@After
	public void tearDown() {
	}

	/**
	 * Test of toSequences method, of class VectorialSequences.
	 */
	@Test
	public void testToSequences_doubleArrArr() {
		//test empty input.
		double[][] arrays = new double[0][10];
		Sequence[] result = VectorialSequences.toSequences(arrays);
		assertEquals(arrays.length, result.length);
		//test correct input.
		arrays = new double[][]{{11, 12, 13}, {21, 22}, {}, {41, 42, 43}};
		result = VectorialSequences.toSequences(arrays);
		assertEquals(arrays.length, result.length);
		for (int s = 0; s < arrays.length; s++) {
			assertEquals(arrays[s].length, result[s].getNodes().size());
			for (int i = 0; i < arrays[s].length; i++) {
				final Node node = result[s].getNodes().get(i);
				final Value value = node.getValue(0);
				assertTrue(value instanceof VectorialValue);
				final double[] vector = ((VectorialValue) value).getVector();
				assertEquals(1, vector.length);
				assertEquals(arrays[s][i], vector[0], 0);
			}
		}
	}

	/**
	 * Test of toSequences method, of class VectorialSequences.
	 */
	@Test
	public void testToSequences_doubleArrArrArr() {
		//test empty input.
		double[][][] arrays = new double[0][10][20];
		Sequence[] result = VectorialSequences.toSequences(arrays);
		assertEquals(arrays.length, result.length);
		//test invalid input
		//empty node.
		arrays = new double[1][1][];
		boolean invalid = false;
		try {
			result = VectorialSequences.toSequences(arrays);
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("Node 0 of sequence 0 is empty!", ex.getMessage());
		}
		assertTrue(invalid);

		invalid = false;
		//inconsistent node dimensionality.
		arrays = new double[][][]{{{1}, {2, 3}}};
		try {
			result = VectorialSequences.toSequences(arrays);
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("Length of node 1 in sequence 0 was not consistent "
					+ "with the rest of the data set!", ex.getMessage());
		}
		assertTrue(invalid);

		//finally test valid sequence conversion.
		arrays = new double[3][][];
		//first sequence:
		arrays[0] = new double[][]{{110, 111}, {120, 121}};
		//second sequence:
		arrays[1] = new double[0][2];
		//third sequence:
		arrays[2] = new double[][]{{310, 311}};
		result = VectorialSequences.toSequences(arrays);
		assertEquals(arrays.length, result.length);
		for (int s = 0; s < arrays.length; s++) {
			assertEquals(arrays[s].length, result[s].getNodes().size());
			for (int i = 0; i < arrays[s].length; i++) {
				final Node node = result[s].getNodes().get(i);
				final Value value = node.getValue(0);
				assertTrue(value instanceof VectorialValue);
				final double[] vector = ((VectorialValue) value).getVector();
				assertEquals(arrays[s][i].length, vector.length);
				assertArrayEquals(arrays[s][i], vector, 0);
			}
		}
	}

	/**
	 * Test of toSequences method, of class VectorialSequences.
	 */
	@Test
	public void testToSequences_doubleArrArrArr_StringArr() {
		//test empty input.
		double[][][] arrays = new double[0][10][20];
		Sequence[] result = VectorialSequences.toSequences(arrays, new String[]{"test"});
		assertEquals(arrays.length, result.length);
		//test invalid input
		//empty node.
		arrays = new double[1][1][];
		boolean invalid = false;
		try {
			result = VectorialSequences.toSequences(arrays, new String[]{"test"});
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("Node 0 of sequence 0 is empty!", ex.getMessage());
		}
		assertTrue(invalid);

		invalid = false;
		//inconsistent node dimensionality.
		arrays = new double[][][]{{{1}, {2, 3}}};
		try {
			result = VectorialSequences.toSequences(arrays, new String[]{"test"});
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("Length of node 1 in sequence 0 was not consistent "
					+ "with the length of the keyword array!", ex.getMessage());
		}
		assertTrue(invalid);

		//finally test valid sequence conversion.
		arrays = new double[3][][];
		//first sequence:
		arrays[0] = new double[][]{{110, 111}, {120, 121}};
		//second sequence:
		arrays[1] = new double[0][2];
		//third sequence:
		arrays[2] = new double[][]{{310, 311}};
		final String[] keywords = {"alpha", "beta"};
		result = VectorialSequences.toSequences(arrays, keywords);
		assertEquals(arrays.length, result.length);
		for (int s = 0; s < arrays.length; s++) {
			assertEquals(arrays[s].length, result[s].getNodes().size());
			for (int i = 0; i < arrays[s].length; i++) {
				final Node node = result[s].getNodes().get(i);
				for (int k = 0; k < keywords.length; k++) {
					final Value value = node.getValue(keywords[k]);
					assertEquals(value, node.getValue(k));
					assertTrue(value instanceof VectorialValue);
					final double[] vector = ((VectorialValue) value).getVector();
					assertEquals(1, vector.length);
					assertEquals(arrays[s][i][k], vector[0], 0);
				}
			}
		}
	}

	/**
	 * Test of toSequences method, of class VectorialSequences.
	 */
	@Test
	public void testToSequences_3args() {
		//test empty input.
		double[][][] arrays = new double[0][10][20];
		Sequence[] result = VectorialSequences.toSequences(arrays, new String[]{"test"}, new int[0]);
		assertEquals(arrays.length, result.length);
		//test invalid input
		//empty node.
		arrays = new double[1][1][];
		boolean invalid = false;
		try {
			result = VectorialSequences.toSequences(arrays, new String[]{"test"});
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("Node 0 of sequence 0 is empty!", ex.getMessage());
		}
		assertTrue(invalid);

		//inconsistent node dimensionality.
		invalid = false;
		arrays = new double[][][]{{{1}, {2, 3}}};
		try {
			result = VectorialSequences.toSequences(arrays, new String[]{"test"}, new int[0]);
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("Length of node 1 in sequence 0 was not consistent "
					+ "with the rest of the data set!", ex.getMessage());
		}
		assertTrue(invalid);

		//too many property dist indices
		invalid = false;
		arrays = new double[][][]{{{1, 2}, {2, 3}}};
		try {
			result = VectorialSequences.toSequences(arrays, new String[]{"test"}, new int[]{8});
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("1 different keywords given but the propertyDist array "
					+ "had 1 entries! We expected 0 entries!", ex.getMessage());
		}
		assertTrue(invalid);

		//too small property dist index
		invalid = false;
		arrays = new double[][][]{{{1, 2}, {2, 3}}};
		try {
			result = VectorialSequences.toSequences(arrays, new String[]{"test", "test2"}, new int[]{0});
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("The entries of the propertyDist array must be larger "
					+ "zero!", ex.getMessage());
		}
		assertTrue(invalid);

		//non-monotonous property distribution
		invalid = false;
		arrays = new double[][][]{{{1, 5, 4}, {2, 3, 2}}};
		try {
			result = VectorialSequences.toSequences(arrays, new String[]{"test", "test2", "test3"}, new int[]{2, 1});
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("The entries of the propertyDist array must increasing!",
					ex.getMessage());
		}
		assertTrue(invalid);

		//too large property dist index
		invalid = false;
		arrays = new double[][][]{{{1, 2}, {2, 3}}};
		try {
			result = VectorialSequences.toSequences(arrays, new String[]{"test", "test2"}, new int[]{8});
		} catch (UnsupportedOperationException ex) {
			invalid = true;
			assertEquals("The entries of the "
					+ "propertyDist array must be smaller than the number "
					+ "of dimensions, which is 2 in this case!", ex.getMessage());
		}
		assertTrue(invalid);

		//finally test valid sequence conversion.
		arrays = new double[3][][];
		//first sequence:
		arrays[0] = new double[][]{{0, 0, 0, 0, 1}, {1, 0, 0, 30, 0}};
		//second sequence:
		arrays[1] = new double[0][];
		//third sequence:
		arrays[2] = new double[][]{{0, 0, 0, 30, 0}};
		final String[] keywords = {"pos", "angle", "speed"};
		final int[] distribution = {3, 4};
		result = VectorialSequences.toSequences(arrays, keywords, distribution);
		assertEquals(arrays.length, result.length);
		for (int s = 0; s < arrays.length; s++) {
			assertEquals(arrays[s].length, result[s].getNodes().size());
			for (int i = 0; i < arrays[s].length; i++) {
				final Node node = result[s].getNodes().get(i);
				//check position
				for (int k = 0; k < keywords.length; k++) {
					final Value value = node.getValue(keywords[k]);
					assertEquals(value, node.getValue(k));
					assertTrue(value instanceof VectorialValue);
					final double[] vector = ((VectorialValue) value).getVector();
					switch (k) {
						//check position:
						case 0:
							assertEquals(3, vector.length);
							assertArrayEquals(
									Arrays.copyOfRange(arrays[s][i], 0, 3),
									vector, 0);
							break;
						default:
							//check single values.
							assertEquals(1, vector.length);
							assertEquals(arrays[s][i][2 + k], vector[0], 0);
					}
				}
			}
		}
	}

	/**
	 * Test of getExtensions method, of class VectorialSequences.
	 */
	@Test
	public void testGetExtensions() {
		//create some extensions using StrictAlignmentFullAlgorithm.
		final int numSamples = 100;
		final double f1 = 0.1;
		final double f2 = 0.07;
		final int K = 2;
		//create two sine waves with two features each different frequencies.
		final double[][][] waves = new double[2][numSamples][K];
		final String[] labels = {"A", "B"};
		for (int i = 0; i < numSamples; i++) {
			waves[0][i][0] = Math.sin(Math.PI * f1 * i);
			waves[0][i][0] = Math.sin(Math.PI * f1 * (i + 1));
			waves[1][i][1] = Math.sin(Math.PI * f2 * i);
			waves[1][i][1] = Math.sin(Math.PI * f2 * (i - 1));
		}
		//transform them to sequences.
		final Sequence[] seqs = VectorialSequences.toSequences(waves, labels);
		//create AlignmentSpecification.
		final AlignmentSpecification spec = VectorialSequences.setUpSpecification(seqs);
		//set up algorithm.
		final StrictAlignmentFullAlgorithm algo = new StrictAlignmentFullAlgorithm(spec);
		//look at alignment path.
		final AlignmentPath path = algo.calculateAlignment(seqs[0], seqs[1]);
		//compare with extensions.
		final double[][][] extensions = VectorialSequences.getExtensions(path);
		for (int i = 0; i < path.getOperations().size(); i++) {
			final Operation op = path.getOperations().get(i);
			if (op.getLeft() == null) {
				assertNull(extensions[0][i]);
			} else {
				assertEquals(K, extensions[0][i].length);
				for (int k = 0; k < K; k++) {
					double[] vec = ((VectorialValue) op.getLeft().getValue(k)).getVector();
					assertEquals(vec[0], extensions[0][i][k], 0);
				}
			}
			if (op.getRight() == null) {
				assertNull(extensions[1][i]);
			} else {
				assertEquals(K, extensions[1][i].length);
				for (int k = 0; k < K; k++) {
					double[] vec = ((VectorialValue) op.getRight().getValue(k)).getVector();
					assertEquals(vec[0], extensions[1][i][k], 0);
				}
			}
		}
	}

}
